# -*- coding: utf-8 -*-
import os

from qiniu.config import _BLOCK_SIZE, get_default

from qiniu.auth import Auth
from qiniu.utils import crc32, file_crc32, rfc_from_timestamp, deprecated

from qiniu.services.storage.uploaders import FormUploader, ResumeUploaderV1, ResumeUploaderV2
from qiniu.services.storage.upload_progress_recorder import UploadProgressRecorder

# for compat to old sdk (<= v7.11.1)
from qiniu.services.storage.legacy import _Resume  # noqa


def put_data(
        up_token,
        key,
        data,
        params=None,
        mime_type='application/octet-stream',
        check_crc=False,
        progress_handler=None,
        fname=None,
        hostscache_dir=None,
        metadata=None,
        regions=None,
        accelerate_uploading=False
):
    """上传二进制流到七牛

    Args:
        up_token:         上传凭证
        key:              上传文件名
        data:             上传二进制流
        params:           自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:        上传数据的mimeType
        check_crc:        是否校验crc32
        progress_handler: 上传进度
        fname:            文件名
        hostscache_dir:   host请求 缓存文件保存位置
        metadata:         元数据
        regions:          区域信息，默认自动查询
        accelerate_uploading: 是否优先使用加速上传

    Returns:
        一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
        一个ResponseInfo对象
    """
    final_data = b''
    if hasattr(data, 'read'):
        while True:
            tmp_data = data.read(_BLOCK_SIZE)
            if len(tmp_data) == 0:
                break
            else:
                final_data += tmp_data
    else:
        final_data = data

    crc = crc32(final_data)
    return _form_put(
        up_token=up_token, key=key, data=final_data, params=params, mime_type=mime_type,
        crc=crc, progress_handler=progress_handler, file_name=fname, metadata=metadata,
        regions=regions, accelerate_uploading=accelerate_uploading
    )


@deprecated("use put_file_v2 instead")
def put_file(
        up_token, key, file_path, params=None,
        mime_type='application/octet-stream', check_crc=False,
        progress_handler=None, upload_progress_recorder=None, keep_last_modified=False, hostscache_dir=None,
        part_size=None, version='v1', bucket_name=None, metadata=None,
        regions=None, accelerate_uploading=False
):
    """上传文件到七牛，此接口的分片传接口默认使用 V1，推荐使用 V2，V2 上传效率更高；在一些专有云服务中需要确认服务是否支持 V2。

        Args:
            up_token:                 上传凭证
            key:                      上传文件名
            file_path:                上传文件的路径
            params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
            mime_type:                上传数据的mimeType
            check_crc:                是否校验crc32
            progress_handler:         上传进度
            upload_progress_recorder: 记录上传进度，用于断点续传
            keep_last_modified:       是否保留文件的最后修改时间
            hostscache_dir:           host请求 缓存文件保存位置，已废弃
            version:                  分片上传版本 目前支持v1/v2版本 默认v1
            part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
            bucket_name:              分片上传v2字段 空间名称
            metadata:                 元数据信息
            regions:                  region信息
            accelerate_uploading:     是否开启加速上传

        Returns:
            一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
            一个ResponseInfo对象
    """
    return _put_file(
        up_token=up_token, key=key, file_path=file_path, params=params, mime_type=mime_type,
        check_crc=check_crc, progress_handler=progress_handler, upload_progress_recorder=upload_progress_recorder,
        keep_last_modified=keep_last_modified, part_size=part_size, version=version, bucket_name=bucket_name,
        metadata=metadata, regions=regions, accelerate_uploading=accelerate_uploading
    )


def put_file_v2(
        up_token, key, file_path, params=None,
        mime_type='application/octet-stream', check_crc=False,
        progress_handler=None, upload_progress_recorder=None, keep_last_modified=False,
        part_size=None, version='v2', bucket_name=None, metadata=None,
        regions=None, accelerate_uploading=False
):
    """上传文件到七牛，此接口的分片传接口默认使用 V2，V2 上传效率更高；在一些专有云服务中需要确认服务是否支持 V2。

        Args:
            up_token:                 上传凭证
            key:                      上传文件名
            file_path:                上传文件的路径
            params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
            mime_type:                上传数据的mimeType
            check_crc:                是否校验crc32
            progress_handler:         上传进度
            upload_progress_recorder: 记录上传进度，用于断点续传
            keep_last_modified:       是否保留文件的最后修改时间
            version:                  分片上传版本 目前支持v1/v2版本 默认v1
            part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
            bucket_name:              分片上传v2字段 空间名称
            metadata:                 元数据信息
            regions:                  region信息
            accelerate_uploading:     是否开启加速上传

        Returns:
            一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
            一个ResponseInfo对象
    """
    return _put_file(
        up_token=up_token, key=key, file_path=file_path, params=params, mime_type=mime_type,
        check_crc=check_crc, progress_handler=progress_handler, upload_progress_recorder=upload_progress_recorder,
        keep_last_modified=keep_last_modified, part_size=part_size, version=version, bucket_name=bucket_name,
        metadata=metadata, regions=regions, accelerate_uploading=accelerate_uploading
    )


def _put_file(
        up_token, key, file_path, params=None,
        mime_type='application/octet-stream', check_crc=False,
        progress_handler=None, upload_progress_recorder=None, keep_last_modified=False,
        part_size=None, version=None, bucket_name=None, metadata=None,
        regions=None, accelerate_uploading=False
):
    """上传文件到七牛

    Args:
        up_token:                 上传凭证
        key:                      上传文件名
        file_path:                上传文件的路径
        params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:                上传数据的mimeType
        check_crc:                是否校验crc32
        progress_handler:         上传进度
        upload_progress_recorder: 记录上传进度，用于断点续传
        keep_last_modified:       是否保留文件的最后修改时间
        version:                  分片上传版本 目前支持v1/v2版本 默认v1
        part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
        bucket_name:              分片上传v2字段 空间名称
        metadata:                 元数据信息
        regions:                  region信息
        accelerate_uploading:     是否开启加速上传

    Returns:
        一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
        一个ResponseInfo对象
    """
    ret = {}
    size = os.stat(file_path).st_size
    with open(file_path, 'rb') as input_stream:
        file_name = os.path.basename(file_path)
        modify_time = int(os.path.getmtime(file_path))
        if size > get_default('default_upload_threshold'):
            ret, info = put_stream_v2(
                up_token=up_token, key=key, input_stream=input_stream, file_name=file_name, data_size=size, params=params,
                mime_type=mime_type, progress_handler=progress_handler,
                upload_progress_recorder=upload_progress_recorder,
                modify_time=modify_time, keep_last_modified=keep_last_modified,
                part_size=part_size, version=version, bucket_name=bucket_name, metadata=metadata,
                regions=regions, accelerate_uploading=accelerate_uploading
            )
        else:
            crc = file_crc32(file_path)
            ret, info = _form_put(
                up_token=up_token, key=key, data=input_stream, params=params, mime_type=mime_type,
                crc=crc, progress_handler=progress_handler, file_name=file_name,
                modify_time=modify_time, keep_last_modified=keep_last_modified, metadata=metadata,
                regions=regions, accelerate_uploading=accelerate_uploading
            )
    return ret, info


def _form_put(
        up_token,
        key,
        data,
        params,
        mime_type,
        crc,
        progress_handler=None,
        file_name=None,
        modify_time=None,
        keep_last_modified=False,
        metadata=None,
        regions=None,
        accelerate_uploading=False
):
    bucket_name = Auth.get_bucket_name(up_token)
    uploader = FormUploader(
        bucket_name,
        progress_handler=progress_handler,
        regions=regions,
        accelerate_uploading=accelerate_uploading,
        preferred_scheme=get_default('default_zone').scheme
    )

    if modify_time and keep_last_modified:
        metadata['x-qn-meta-!Last-Modified'] = rfc_from_timestamp(modify_time)

    return uploader.upload(
        key=key,
        data=data,
        file_name=file_name,
        modify_time=modify_time,
        mime_type=mime_type,
        metadata=metadata,
        custom_vars=params,
        crc32_int=crc,
        up_token=up_token
    )


@deprecated("use put_stream_v2 instead")
def put_stream(
        up_token,
        key,
        input_stream,
        file_name,
        data_size,
        hostscache_dir=None,
        params=None,
        mime_type=None,
        progress_handler=None,
        upload_progress_recorder=None,
        modify_time=None,
        keep_last_modified=False,
        part_size=None,
        version='v1',
        bucket_name=None,
        metadata=None,
        regions=None,
        accelerate_uploading=False
):
    """ 通过 stream 方式上传文件到七牛，此接口的分片传接口默认使用 V1，推荐使用 V2，V2 上传效率更高；在一些专有云服务中需要确认服务是否支持 V2。

    Args:
        up_token:                 上传凭证
        key:                      上传文件名
        input_stream:             上传数据流
        file_name:                文件名
        data_size:                数据流大小
        hostscache_dir:           host请求 缓存文件保存位置，当前已废弃
        params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
        mime_type:                上传数据的mimeType
        progress_handler:         上传进度
        upload_progress_recorder: 记录上传进度，用于断点续传
        modify_time:              数据修改时间
        keep_last_modified:       是否保留文件的最后修改时间
        part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
        version:                  分片上传版本 目前支持v1/v2版本 默认v1
        bucket_name:              分片上传v2字段 空间名称
        metadata:                 元数据信息
        regions:                  region信息
        accelerate_uploading:     是否开启加速上传

    Returns:
        一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
        一个ResponseInfo对象
    """
    return _put_stream(
        up_token=up_token,
        key=key,
        input_stream=input_stream,
        file_name=file_name,
        data_size=data_size,
        params=params,
        mime_type=mime_type,
        progress_handler=progress_handler,
        upload_progress_recorder=upload_progress_recorder,
        modify_time=modify_time,
        keep_last_modified=keep_last_modified,
        part_size=part_size,
        version=version,
        bucket_name=bucket_name,
        metadata=metadata,
        regions=regions,
        accelerate_uploading=accelerate_uploading
    )


def put_stream_v2(
        up_token,
        key,
        input_stream,
        file_name,
        data_size,
        params=None,
        mime_type=None,
        progress_handler=None,
        upload_progress_recorder=None,
        modify_time=None,
        keep_last_modified=False,
        part_size=None,
        version='v2',
        bucket_name=None,
        metadata=None,
        regions=None,
        accelerate_uploading=False
):
    """ 通过 stream 方式上传文件到七牛，此接口的分片传接口默认使用 V2，V2 上传效率更高；在一些专有云服务中需要确认服务是否支持 V2。

        Args:
            up_token:                 上传凭证
            key:                      上传文件名
            input_stream:             上传数据流
            file_name:                文件名
            data_size:                数据流大小
            params:                   自定义变量，规格参考 https://developer.qiniu.com/kodo/manual/vars#xvar
            mime_type:                上传数据的mimeType
            progress_handler:         上传进度
            upload_progress_recorder: 记录上传进度，用于断点续传
            modify_time:              数据修改时间
            keep_last_modified:       是否保留文件的最后修改时间
            part_size:                分片上传v2必传字段 默认大小为4MB 分片大小范围为1 MB - 1 GB
            version:                  分片上传版本 目前支持v1/v2版本 默认v1
            bucket_name:              分片上传v2字段 空间名称
            metadata:                 元数据信息
            regions:                  region信息
            accelerate_uploading:     是否开启加速上传

        Returns:
            一个dict变量，类似 {"hash": "<Hash string>", "key": "<Key string>"}
            一个ResponseInfo对象
    """
    return _put_stream(
        up_token=up_token,
        key=key,
        input_stream=input_stream,
        file_name=file_name,
        data_size=data_size,
        params=params,
        mime_type=mime_type,
        progress_handler=progress_handler,
        upload_progress_recorder=upload_progress_recorder,
        modify_time=modify_time,
        keep_last_modified=keep_last_modified,
        part_size=part_size,
        version=version,
        bucket_name=bucket_name,
        metadata=metadata,
        regions=regions,
        accelerate_uploading=accelerate_uploading
    )


def _put_stream(
        up_token,
        key,
        input_stream,
        file_name,
        data_size,
        params=None,
        mime_type=None,
        progress_handler=None,
        upload_progress_recorder=None,
        modify_time=None,
        keep_last_modified=False,
        part_size=None,
        version='v2',
        bucket_name=None,
        metadata=None,
        regions=None,
        accelerate_uploading=False
):
    if not bucket_name:
        bucket_name = Auth.get_bucket_name(up_token)
    if not upload_progress_recorder:
        upload_progress_recorder = UploadProgressRecorder()
    if not version:
        version = 'v1'
    if not part_size:
        part_size = 4 * (1024 * 1024)

    if version == 'v1':
        uploader = ResumeUploaderV1(
            bucket_name,
            progress_handler=progress_handler,
            upload_progress_recorder=upload_progress_recorder,
            regions=regions,
            accelerate_uploading=accelerate_uploading,
            preferred_scheme=get_default('default_zone').scheme
        )
        if modify_time and keep_last_modified:
            metadata['x-qn-meta-!Last-Modified'] = rfc_from_timestamp(modify_time)
    elif version == 'v2':
        uploader = ResumeUploaderV2(
            bucket_name,
            progress_handler=progress_handler,
            upload_progress_recorder=upload_progress_recorder,
            part_size=part_size,
            regions=regions,
            accelerate_uploading=accelerate_uploading,
            preferred_scheme=get_default('default_zone').scheme
        )
    else:
        raise ValueError('version only could be v1 or v2')
    return uploader.upload(
        key=key,
        data=input_stream,
        data_size=data_size,
        file_name=file_name,
        modify_time=modify_time,
        mime_type=mime_type,
        metadata=metadata,
        custom_vars=params,
        up_token=up_token
    )



from qiniu.http.endpoints_retry_policy import EndpointsRetryPolicy
from qiniu.http.regions_retry_policy import RegionsRetryPolicy
from qiniu.retry import Retrier


def get_default_retrier(
    regions_provider,
    service_names,
    preferred_endpoints_provider=None,
):
    if not service_names:
        raise ValueError('service_names should not be empty')

    retry_policies = [
        EndpointsRetryPolicy(
            skip_init_context=True
        ),
        RegionsRetryPolicy(
            regions_provider=regions_provider,
            service_names=service_names,
            preferred_endpoints_provider=preferred_endpoints_provider
        )
    ]

    return Retrier(retry_policies)



import functools
import math
from collections import namedtuple
from concurrent import futures
from io import BytesIO
from os import path
from threading import Lock
from time import time

from qiniu.compat import is_seekable
from qiniu.auth import Auth
from qiniu.http import qn_http_client, ResponseInfo
from qiniu.http.endpoint import Endpoint
from qiniu.utils import b, io_md5, urlsafe_base64_encode
from qiniu.compat import json

from ._default_retrier import ProgressRecord, get_default_retrier
from .abc import ResumeUploaderBase
from .io_chunked import IOChunked


class ResumeUploaderV2(ResumeUploaderBase):
    def _recover_from_record(
        self,
        file_name,
        key,
        context
    ):
        """

        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context

        Returns
        -------
        _ResumeUploadV2Context
        """
        if not isinstance(context, _ResumeUploadV2Context):
            raise TypeError('"context" must be an instance of _ResumeUploadV2Context')

        if (
            not self.upload_progress_recorder or
            not (file_name or key)
        ):
            return context

        record = self.upload_progress_recorder.get_upload_record(
            file_name,
            key
        )

        if not record:
            return context

        record_up_hosts = record.get('up_hosts', [])
        record_upload_id = record.get('upload_id', '')
        record_expired_at = record.get('expired_at', 0)
        record_part_size = record.get('part_size', None)
        record_modify_time = record.get('modify_time', 0)
        record_etags = record.get('etags', [])

        # compat with old sdk(<= v7.11.1)
        if not record_up_hosts or not record_part_size:
            return context

        if (
            not record_upload_id or
            record_modify_time != context.modify_time or
            record_expired_at < time()
        ):
            return context

        return context._replace(
            up_hosts=record_up_hosts,
            upload_id=record_upload_id,
            expired_at=record_expired_at,
            part_size=record_part_size,
            parts=[
                _ResumeUploadV2Part(
                    part_no=p['partNumber'],
                    etag=p['etag']
                )
                for p in record_etags
                if (
                    p.get('partNumber', None) and
                    p.get('etag', None)
                )
            ],
            resumed=True
        )

    def _set_to_record(self, file_name, key, context):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context

        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return

        record_data = {
            'up_hosts': context.up_hosts,
            'upload_id': context.upload_id,
            'expired_at': context.expired_at,
            'part_size': context.part_size,
            'modify_time': context.modify_time,
            'etags': [
                {
                    'etag': part.etag,
                    'partNumber': part.part_no
                }
                for part in context.parts
            ]
        }
        self.upload_progress_recorder.set_upload_record(
            file_name,
            key,
            data=record_data
        )

    def _try_delete_record(
        self,
        file_name,
        key,
        context=None,
        resp=None
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context
        resp: ResponseInfo
        """
        if not self.upload_progress_recorder or not any([file_name, key]):
            return
        if resp and not resp.ok():
            return
        self.upload_progress_recorder.delete_upload_record(file_name, key)

    def _progress_handler(
        self,
        file_name,
        key,
        context,
        uploaded_size,
        total_size
    ):
        """
        Parameters
        ----------
        file_name: str
        key: str
        context: _ResumeUploadV2Context
        uploaded_size: int
        total_size: int
        """
        self._set_to_record(file_name, key, context)
        if not callable(self.progress_handler):
            return
        try:
            self.progress_handler(uploaded_size, total_size)
        except Exception as err:
            err.no_need_retry = True
            raise err

    def _initial_context(
        self,
        key,
        file_name,
        modify_time,
        part_size
    ):
        context = _ResumeUploadV2Context(
            up_hosts=[],
            upload_id='',
            expired_at=0,
            part_size=part_size,
            parts=[],
            modify_time=modify_time,
            resumed=False
        )

        # try to recover from record

        return self._recover_from_record(
            file_name,
            key,
            context
        )

    def initial_parts(
        self,
        up_token,
        key,
        file_path=None,
        data=None,
        data_size=None,
        modify_time=None,
        part_size=None,
        file_name=None,
        up_endpoint=None,
        **kwargs
    ):
        """

        Parameters
        ----------
        up_token: str
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int
        part_size: int
        file_name: str
        up_endpoint: Endpoint
        kwargs

        Returns
        -------
        ret: _ResumeUploadV2Context
        resp: ResponseInfo
        """
        # -- check and initial arguments
        # must provide file_path or data
        if not file_path and not data:
            raise TypeError('Must provide one of file_path or data.')
        if file_path and data:
            raise TypeError('Must provide only one of file_path or data.')

        # data must has length
        if not file_path and not data_size:
            raise TypeError('Must provide size if use data.')

        if not modify_time:
            if file_path:
                modify_time = int(path.getmtime(file_path))
            else:
                modify_time = int(time())

        if not part_size:
            part_size = self.part_size

        # -- initial context
        if not file_name and file_path:
            file_name = path.basename(file_path)
        context = self._initial_context(
            key=key,
            file_name=file_name,
            modify_time=modify_time,
            part_size=part_size
        )

        if (
            context.up_hosts and
            context.upload_id and
            context.expired_at
        ):
            return context, None

        # -- get a new upload id
        if not context.up_hosts and up_endpoint:
            context.up_hosts.extend([up_endpoint.get_value(scheme=self.preferred_scheme)])

        if not context.up_hosts:
            access_key, _, _ = Auth.up_token_decode(up_token)
            context.up_hosts.extend(self._get_up_hosts(access_key))

        bucket_name = Auth.get_bucket_name(up_token)

        resp = None
        for up_host in context.up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key
            )
            ret, resp = qn_http_client.post(
                url=url,
                data='',
                files=None,
                headers={
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if not resp.ok() and not resp.need_retry():
                break
            if resp.ok() and ret:
                context = context._replace(
                    upload_id=ret.get('uploadId', ''),
                    expired_at=ret.get('expireAt', 0)
                )
                break

        return context, resp

    def upload_parts(
        self,
        up_token,
        data,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data
        data_size: int
        context: _ResumeUploadV2Context
        kwargs
            key, file_name

        Returns
        -------
        part: _ResumeUploadV2Part
        resp: ResponseInfo

        """
        # initial arguments
        chunk_list = self.gen_chunk_list(
            size=data_size,
            chunk_size=context.part_size,
            uploaded_chunk_no_list=[
                p.part_no for p in context.parts
            ]
        )
        up_hosts = list(context.up_hosts)
        file_name = kwargs.get('file_name', None)
        key = kwargs.get('key', None)

        # initial upload state
        part, resp = None, None
        uploaded_size = context.part_size * len(context.parts)
        if math.ceil(data_size / context.part_size) in [p.part_no for p in context.parts]:
            # if last part uploaded, should correct the uploaded size
            uploaded_size += (data_size % context.part_size) - context.part_size
        lock = Lock()

        if not self.concurrent_executor:
            # upload sequentially
            for chunk in chunk_list:
                part, resp = self.__upload_part(
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    upload_id=context.upload_id,
                    key=key,
                    lock=lock
                )
                if not resp.ok():
                    return None, resp
                context.parts.append(part)
                uploaded_size += chunk.chunk_size
                self._progress_handler(
                    file_name=file_name,
                    key=key,
                    context=context,
                    uploaded_size=uploaded_size,
                    total_size=data_size
                )
        else:
            # upload concurrently
            future_chunk_dict = {}
            for chunk in chunk_list:
                ftr = self.concurrent_executor.submit(
                    self.__upload_part,
                    data=data,
                    chunk_info=chunk,
                    up_hosts=up_hosts,
                    up_token=up_token,
                    upload_id=context.upload_id,
                    key=key,
                    lock=lock
                )
                future_chunk_dict[ftr] = chunk

            first_failed_resp = None
            for ftr in futures.as_completed(future_chunk_dict):
                if ftr.cancelled():
                    continue
                elif ftr.exception():
                    if first_failed_resp:
                        continue
                    first_failed_resp = ResponseInfo(None, ftr.exception())
                    for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                        not_done.cancel()
                else:
                    part, resp = ftr.result()
                    if not part:
                        if not first_failed_resp:
                            first_failed_resp = resp
                            for not_done in filter(lambda f: not f.done(), future_chunk_dict):
                                not_done.cancel()
                    else:
                        context.parts.append(part)
                        uploaded_size += future_chunk_dict[ftr].chunk_size
                        self._progress_handler(
                            file_name=file_name,
                            key=key,
                            context=context,
                            uploaded_size=uploaded_size,
                            total_size=data_size
                        )
            if first_failed_resp:
                return None, first_failed_resp

        return part, resp

    def complete_parts(
        self,
        up_token,
        data_size,
        context,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        data_size: int
        context: _ResumeUploadV2Context
        kwargs
            key, file_name, params, metadata
        Returns
        -------
            ret: dict
            resp: ResponseInfo
        """
        key = kwargs.get('key', None)
        file_name = kwargs.get('file_name', None)
        mime_type = kwargs.get('mime_type', None)
        params = kwargs.get('params', None)
        metadata = kwargs.get('metadata', None)

        # sort contexts
        sorted_parts = sorted(context.parts, key=lambda part: part.part_no)

        bucket_name = Auth.get_bucket_name(up_token)

        ret, resp = None, None
        for up_host in context.up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key,
                upload_id=context.upload_id
            )
            data = {
                'parts': [
                    {
                        'etag': p.etag,
                        'partNumber': p.part_no
                    }
                    for p in sorted_parts
                ],
                'fname': file_name,
                'mimeType': mime_type,
                'customVars': params,
                'metadata': metadata
            }
            ret, resp = qn_http_client.post(
                url=url,
                data=json.dumps(data),
                files=None,
                headers={
                    'Content-Type': 'application/json',
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() or not resp.need_retry():
                break
        self._try_delete_record(
            file_name,
            key,
            context,
            resp
        )
        return ret, resp

    def __upload_with_retrier(
        self,
        access_key,
        bucket_name,
        **upload_opts
    ):
        file_name = upload_opts.get('file_name', None)
        key = upload_opts.get('key', None)
        modify_time = upload_opts.get('modify_time', None)
        part_size = upload_opts.get('part_size', self.part_size)

        context = self._initial_context(
            key=key,
            file_name=file_name,
            modify_time=modify_time,
            part_size=part_size
        )
        preferred_endpoints = None
        if context.up_hosts:
            preferred_endpoints = [
                Endpoint.from_host(h)
                for h in context.up_hosts
            ]

        progress_record = None
        if all(
            [
                self.upload_progress_recorder,
                file_name,
                key
            ]
        ):
            progress_record = ProgressRecord(
                upload_api_version='v1',
                exists=functools.partial(
                    self.upload_progress_recorder.has_upload_record,
                    file_name=file_name,
                    key=key
                ),
                delete=functools.partial(
                    self.upload_progress_recorder.delete_upload_record,
                    file_name=file_name,
                    key=key
                )
            )

        retrier = get_default_retrier(
            regions_provider=self._get_regions_provider(
                access_key=access_key,
                bucket_name=bucket_name
            ),
            preferred_endpoints_provider=preferred_endpoints,
            progress_record=progress_record,
            accelerate_uploading=self.accelerate_uploading
        )

        data = upload_opts.get('data')
        attempt = None
        for attempt in retrier:
            with attempt:
                upload_opts['up_endpoint'] = attempt.context.get('endpoint')
                attempt.result = self.__upload(
                    **upload_opts
                )
                ret, resp = attempt.result
                if resp.ok() and ret:
                    return attempt.result
                if (
                    not is_seekable(data) or
                    not resp.need_retry()
                ):
                    return attempt.result
                data.seek(0)

        if attempt is None:
            raise RuntimeError('Retrier is not working. attempt is None')

        return attempt.result

    def __upload(
        self,
        up_token,
        key,
        file_path,
        file_name,
        data,
        data_size,
        part_size,
        modify_time,
        mime_type,
        custom_vars,
        metadata,
        up_endpoint
    ):
        # initial_parts
        context, resp = self.initial_parts(
            up_token,
            key,
            file_path=file_path,
            file_name=file_name,
            data=data,
            data_size=data_size,
            modify_time=modify_time,
            part_size=part_size,
            up_endpoint=up_endpoint
        )

        if (
            not context.up_hosts or
            not context.upload_id or
            not context.expired_at
        ):
            return None, resp

        # upload parts
        try:
            if file_path:
                data_size = path.getsize(file_path)
                data = open(file_path, 'rb')
            elif isinstance(data, bytes):
                data_size = len(data)
                data = BytesIO(data)
            elif isinstance(data, str):
                data_size = len(data)
                data = BytesIO(b(data))
            ret, resp = self.upload_parts(
                up_token=up_token,
                data=data,
                data_size=data_size,
                context=context,

                key=key,
                file_name=file_name
            )
        finally:
            if file_path:
                data.close()

        if resp and not resp.ok():
            return ret, resp

        # complete parts
        ret, resp = self.complete_parts(
            up_token=up_token,
            data_size=data_size,
            context=context,

            key=key,
            mime_type=mime_type,
            file_name=file_name,
            params=custom_vars,
            metadata=metadata
        )

        return ret, resp

    def upload(
        self,
        key,
        file_path=None,
        data=None,
        data_size=None,

        part_size=None,
        modify_time=None,
        mime_type=None,
        metadata=None,
        file_name=None,
        custom_vars=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        key: str
        file_path: str
        data: IOBase
        data_size: int
        part_size: int
        modify_time: int
        mime_type: str
        metadata: dict
        file_name: str
        custom_vars: dict
        kwargs
            up_token: str
            bucket_name: str,
            expired: int,
            policy: dict,
            strict_policy: bool

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """
        # up_token
        up_token = kwargs.get('up_token', None)
        if not up_token:
            kwargs.setdefault('up_token', self.get_up_token(**kwargs))
            access_key = self.auth.get_access_key()
        else:
            access_key, _, _ = Auth.up_token_decode(up_token)

        # bucket_name
        kwargs['bucket_name'] = Auth.get_bucket_name(up_token)

        # file_name
        if not file_name and file_path:
            file_name = path.basename(file_path)

        # upload
        return self.__upload_with_retrier(
            access_key=access_key,
            key=key,
            file_path=file_path,
            file_name=file_name,
            data=data,
            data_size=data_size,
            part_size=part_size,
            modify_time=modify_time,
            mime_type=mime_type,
            custom_vars=custom_vars,
            metadata=metadata,
            **kwargs
        )

    def __get_url_for_upload(
        self,
        up_host,
        bucket_name,
        key,
        upload_id=None,
        part_no=None,
    ):
        """
        Parameters
        ----------
        up_host: str
        bucket_name: str
        key: str
        upload_id: str
        part_no: int

        Returns
        -------
        str
        """
        if not bucket_name:
            bucket_name = self.bucket_name

        object_entry = '~'
        if key:
            object_entry = urlsafe_base64_encode(key)

        url_segs = [
            up_host,
            'buckets', bucket_name,
            'objects', object_entry,
            'uploads',
        ]

        if upload_id:
            url_segs.append(upload_id)

        if part_no:
            url_segs.append(str(part_no))

        return '/'.join(url_segs)

    def __upload_part(
        # resort arguments
        self,
        data,
        chunk_info,
        up_hosts,
        up_token,
        upload_id,
        key,
        lock
    ):
        """
        Parameters
        ----------
        data: IOBase
        chunk_info: ChunkInfo
        up_hosts: list[str]
        up_token: str
        upload_id: str
        key: str
        lock: Lock

        Returns
        -------
        part: _ResumeUploadV2Part
        resp: ResponseInfo
        """
        if not up_hosts:
            raise ValueError('Must provide on up host at least')

        bucket_name = Auth.get_bucket_name(up_token)
        if not bucket_name:
            bucket_name = self.bucket_name

        chunked_data = IOChunked(
            base_io=data,
            chunk_offset=chunk_info.chunk_offset,
            chunk_size=chunk_info.chunk_size,
            lock=lock
        )
        chunk_md5 = io_md5(chunked_data)
        chunked_data.seek(0)
        part, resp = None, None
        for up_host in up_hosts:
            url = self.__get_url_for_upload(
                up_host,
                bucket_name,
                key,
                upload_id=upload_id,
                part_no=chunk_info.chunk_no
            )
            ret, resp = qn_http_client.put(
                url=url,
                data=chunked_data,
                files=None,
                headers={
                    'Content-Type': 'application/octet-stream',
                    'Content-MD5': chunk_md5,
                    'Authorization': 'UpToken {}'.format(up_token)
                }
            )
            if resp.ok() and ret:
                part = _ResumeUploadV2Part(
                    part_no=chunk_info.chunk_no,
                    etag=ret.get('etag', '')
                )
                return part, resp
            if (
                not is_seekable(chunked_data) or
                not resp.need_retry()
            ):
                return part, resp
            chunked_data.seek(0)
        return part, resp


# use dataclass instead namedtuple if min version of python update to 3.7
_ResumeUploadV2Part = namedtuple(
    'ResumeUploadV2Part',
    [
        'part_no',
        'etag'
    ]
)

_ResumeUploadV2Context = namedtuple(
    'ResumeUploadV2Context',
    [
        'up_hosts',
        'upload_id',
        'expired_at',
        'part_size',
        'parts',
        'modify_time',
        'resumed'
    ]
)



# -*- coding: utf-8 -*-

from qiniu import http
import json

from qiniu.compat import is_py2
from qiniu.compat import is_py3
from enum import Enum

import hashlib


class DataType(Enum):
    BANDWIDTH = 'bandwidth'
    X302BANDWIDTH = '302bandwidth'
    X302MBANDWIDTH = '302mbandwidth'
    FLOW = 'flow'
    X302FLOW = '302flow'
    X302MFLOW = '302mflow'


def urlencode(str):
    if is_py2:
        import urllib2
        return urllib2.quote(str)
    elif is_py3:
        import urllib.parse
        return urllib.parse.quote(str)


class CdnManager(object):
    def __init__(self, auth):
        self.auth = auth
        self.server = 'http://fusion.qiniuapi.com'

    def refresh_urls(self, urls):
        """
        刷新文件列表，文档 https://developer.qiniu.com/fusion/api/cache-refresh

        Args:
            urls: 待刷新的文件外链列表

        Returns:
            一个dict变量和一个ResponseInfo对象
            参考代码 examples/cdn_manager.py
        """
        return self.refresh_urls_and_dirs(urls, None)

    def refresh_dirs(self, dirs):
        """
        刷新目录，文档 https://developer.qiniu.com/fusion/api/cache-refresh

        Args:
            urls: 待刷新的目录列表

        Returns:
            一个dict变量和一个ResponseInfo对象
            参考代码 examples/cdn_manager.py
        """
        return self.refresh_urls_and_dirs(None, dirs)

    def refresh_urls_and_dirs(self, urls, dirs):
        """
        刷新文件目录，文档 https://developer.qiniu.com/fusion/api/cache-refresh

        Args:
           urls: 待刷新的目录列表
           dirs: 待刷新的文件列表

        Returns:
           一个dict变量和一个ResponseInfo对象
           参考代码 examples/cdn_manager.py
        """
        req = {}
        if urls is not None and len(urls) > 0:
            req.update({"urls": urls})
        if dirs is not None and len(dirs) > 0:
            req.update({"dirs": dirs})

        body = json.dumps(req)
        url = '{0}/v2/tune/refresh'.format(self.server)
        return self.__post(url, body)

    def prefetch_urls(self, urls):
        """
        预取文件列表，文档 https://developer.qiniu.com/fusion/api/file-prefetching

        Args:
           urls: 待预取的文件外链列表

        Returns:
           一个dict变量和一个ResponseInfo对象
           参考代码 examples/cdn_manager.py
        """
        req = {}
        req.update({"urls": urls})

        body = json.dumps(req)
        url = '{0}/v2/tune/prefetch'.format(self.server)
        return self.__post(url, body)

    def get_bandwidth_data(self, domains, start_date, end_date, granularity, data_type=None):
        """
        查询带宽数据，文档 https://developer.qiniu.com/fusion/api/traffic-bandwidth

        Args:
           domains:     域名列表
           start_date:  起始日期
           end_date:    结束日期
           granularity: 数据间隔
           data_type:   计量数据类型, see class DataType.XXXBANDWIDTH

        Returns:
           一个dict变量和一个ResponseInfo对象
           参考代码 examples/cdn_manager.py
        """
        req = {}
        req.update({"domains": ';'.join(domains)})
        req.update({"startDate": start_date})
        req.update({"endDate": end_date})
        req.update({"granularity": granularity})
        if data_type is not None:
            req.update({'type': data_type.value})  # should be one of 'bandwidth', '302bandwidth', '302mbandwidth'

        body = json.dumps(req)
        url = '{0}/v2/tune/bandwidth'.format(self.server)
        return self.__post(url, body)

    def get_flux_data(self, domains, start_date, end_date, granularity, data_type=None):
        """
        查询流量数据，文档 https://developer.qiniu.com/fusion/api/traffic-bandwidth

        Args:
           domains:     域名列表
           start_date:  起始日期
           end_date:    结束日期
           granularity: 数据间隔
           data_type:   计量数据类型, see class DataType.XXXFLOW

        Returns:
           一个dict变量和一个ResponseInfo对象
           参考代码 examples/cdn_manager.py
        """
        req = {}
        req.update({"domains": ';'.join(domains)})
        req.update({"startDate": start_date})
        req.update({"endDate": end_date})
        req.update({"granularity": granularity})
        if data_type is not None:
            req.update({'type': data_type.value})  # should be one of 'flow', '302flow', '302mflow'

        body = json.dumps(req)
        url = '{0}/v2/tune/flux'.format(self.server)
        return self.__post(url, body)

    def get_log_list_data(self, domains, log_date):
        """
        获取日志下载链接，文档 https://developer.qiniu.com/fusion/api/download-the-log

        Args:
           domains:     域名列表
           log_date:    日志日期

        Returns:
           一个dict变量和一个ResponseInfo对象
           参考代码 examples/cdn_manager.py
        """
        req = {}
        req.update({"domains": ';'.join(domains)})
        req.update({"day": log_date})

        body = json.dumps(req)
        url = '{0}/v2/tune/log/list'.format(self.server)
        return self.__post(url, body)

    def put_httpsconf(self, name, certid, forceHttps=False):
        """
        修改证书，文档 https://developer.qiniu.com/fusion/4246/the-domain-name#11

        Args:
           domains:     域名name
           CertID:      证书id，从上传或者获取证书列表里拿到证书id
           ForceHttps:  是否强制https跳转

        Returns:
           {}
        """
        req = {}
        req.update({"certid": certid})
        req.update({"forceHttps": forceHttps})

        body = json.dumps(req)
        url = '{0}/domain/{1}/httpsconf'.format(self.server, name)
        return self.__post(url, body)

    def __post(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return http._post_with_auth_and_headers(url, data, self.auth, headers)


class DomainManager(object):
    def __init__(self, auth):
        self.auth = auth
        self.server = 'http://api.qiniu.com'

    def create_domain(self, name, body):
        """
        创建域名，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name

        Args:
           name:     域名, 如果是泛域名，必须以点号 . 开头
           bosy:     创建域名参数
        Returns:
           {}
        """
        url = '{0}/domain/{1}'.format(self.server, name)
        return self.__post(url, body)

    def domain_online(self, name):
        """
        上线域名，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name#6

        Args:
            name:     域名, 如果是泛域名，必须以点号 . 开头
            bosy:     创建域名参数
        Returns:
            {}
        """
        url = '{0}/domain/{1}/online'.format(self.server, name)
        return http._post_with_qiniu_mac(url, None, self.auth)

    def domain_offline(self, name):
        """
        下线域名，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name#5

        Args:
            name:     域名, 如果是泛域名，必须以点号 . 开头
            bosy:     创建域名参数
        Returns:
            {}
        """
        url = '{0}/domain/{1}/offline'.format(self.server, name)
        return http._post_with_qiniu_mac(url, None, self.auth)

    def delete_domain(self, name):
        """
        删除域名，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name#8

        Args:
            name:     域名, 如果是泛域名，必须以点号 . 开头
        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/domain/{1}'.format(self.server, name)
        return self.__del(url)

    def get_domain(self, name):
        """
        获取域名信息，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name

        Args:
           name:     域名, 如果是泛域名，必须以点号 . 开头
        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回dict{}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息
        """
        url = '{0}/domain/{1}'.format(self.server, name)
        return self.__get(url)

    def put_httpsconf(self, name, certid, forceHttps):
        """
        修改证书，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name#11

        Args:
           domains:     域名name
           CertID:      证书id，从上传或者获取证书列表里拿到证书id
           ForceHttps:  是否强制https跳转

        Returns:
           {}
        """
        req = {}
        req.update({"certid": certid})
        req.update({"forceHttps": forceHttps})

        body = json.dumps(req)
        url = '{0}/domain/{1}/httpsconf'.format(self.server, name)
        return self.__put(url, body)

    def create_sslcert(self, name, common_name, pri, ca):
        """
        修改证书，文档 https://developer.qiniu.com/fusion/api/4246/the-domain-name#11

        Args:
           name:        证书名称
           common_name: 相关域名
           pri:         证书私钥
           ca:          证书内容
        Returns:
            返回一个tuple对象，其格式为(<result>, <ResponseInfo>)
            - result          成功返回dict{certID: <CertID>}，失败返回{"error": "<errMsg string>"}
            - ResponseInfo    请求的Response信息


        """
        req = {}
        req.update({"name": name})
        req.update({"common_name": common_name})
        req.update({"pri": pri})
        req.update({"ca": ca})

        body = json.dumps(req)
        url = '{0}/sslcert'.format(self.server)
        return self.__post(url, body)

    def __post(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return http._post_with_auth_and_headers(url, data, self.auth, headers)

    def __put(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return http._put_with_auth_and_headers(url, data, self.auth, headers)

    def __get(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return http._get_with_auth_and_headers(url, data, self.auth, headers)

    def __del(self, url, data=None):
        headers = {'Content-Type': 'application/json'}
        return http._delete_with_qiniu_mac_and_headers(url, data, self.auth, headers)


def create_timestamp_anti_leech_url(host, file_name, query_string, encrypt_key, deadline):
    """
    创建时间戳防盗链

    Args:
        host:              带访问协议的域名
        file_name:         原始文件名，不需要urlencode
        query_string:      查询参数，不需要urlencode
        encrypt_key:       时间戳防盗链密钥
        deadline:          链接有效期时间戳（以秒为单位）

    Returns:
        带时间戳防盗链鉴权访问链接
    """
    if query_string:
        url_to_sign = '{0}/{1}?{2}'.format(host, urlencode(file_name), query_string)
    else:
        url_to_sign = '{0}/{1}'.format(host, urlencode(file_name))

    path = '/{0}'.format(urlencode(file_name))
    expire_hex = str(hex(deadline))[2:]
    str_to_sign = '{0}{1}{2}'.format(encrypt_key, path, expire_hex).encode()
    sign_str = hashlib.md5(str_to_sign).hexdigest()

    if query_string:
        signed_url = '{0}&sign={1}&t={2}'.format(url_to_sign, sign_str, expire_hex)
    else:
        signed_url = '{0}?sign={1}&t={2}'.format(url_to_sign, sign_str, expire_hex)

    return signed_url



from io import BytesIO
from os import path
from time import time

from qiniu.compat import is_seekable
from qiniu.utils import b, io_crc32
from qiniu.auth import Auth
from qiniu.http import qn_http_client

from .abc import UploaderBase
from ._default_retrier import get_default_retrier


class FormUploader(UploaderBase):
    def __init__(self, bucket_name, **kwargs):
        """
        Parameters
        ----------
        bucket_name: str
        kwargs
            auth, regions
        """
        super(FormUploader, self).__init__(bucket_name, **kwargs)

        self.progress_handler = kwargs.get(
            'progress_handler',
            None
        )

    def upload(
        self,
        key,
        file_path=None,
        data=None,
        data_size=None,
        modify_time=None,
        part_size=None,
        mime_type=None,
        metadata=None,
        file_name=None,
        custom_vars=None,
        **kwargs
    ):
        """
        Parameters
        ----------
        key: str
        file_path: str
        data: IOBase
        data_size: int
        modify_time: int
        part_size: int
        mime_type: str
        metadata: dict
        file_name: str
        custom_vars: dict
        kwargs
            up_token: str
            crc32_int: int
            bucket_name: str
                is required if upload to another bucket
            expired: int
                option for generate up_token if not provide up_token. seconds
            policy: dict
                option for generate up_token if not provide up_token. details see `auth.Auth`
            strict_policy: bool
                option for generate up_token if not provide up_token

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """
        # check and initial arguments
        # bucket_name
        bucket_name = kwargs.get('bucket_name', self.bucket_name)

        # up_token
        up_token = kwargs.get('up_token', None)
        if not up_token:
            up_token = self.get_up_token(**kwargs)
            access_key = self.auth.get_access_key()
        else:
            access_key, _, _ = Auth.up_token_decode(up_token)

        # crc32 from outside
        crc32_int = kwargs.get('crc32_int', None)
        # try to get file_name
        if not file_name and file_path:
            file_name = path.basename(file_path)

        # must provide file_path or data
        if not file_path and not data:
            raise TypeError('Must provide one of file_path or data.')
        if file_path and data:
            raise TypeError('Must provide only one of file_path or data.')

        # useless for form upload
        if not modify_time:
            if file_path:
                modify_time = int(path.getmtime(file_path))
            else:
                modify_time = int(time())

        # upload
        try:
            if file_path:
                data_size = path.getsize(file_path)
                data = open(file_path, 'rb')
            elif isinstance(data, bytes):
                data_size = len(data)
                data = BytesIO(data)
            elif isinstance(data, str):
                data_size = len(data)
                data = BytesIO(b(data))
            if not crc32_int:
                crc32_int = self.__get_crc32_int(data)
            fields = self.__get_form_fields(
                up_token=up_token,
                key=key,
                crc32_int=crc32_int,
                custom_vars=custom_vars,
                metadata=metadata
            )
            ret, resp = self.__upload_data_with_retrier(
                # retrier options
                access_key=access_key,
                bucket_name=bucket_name,
                # upload_data options
                fields=fields,
                file_name=file_name,
                data=data,
                data_size=data_size,
                mime_type=mime_type
            )
        finally:
            if file_path:
                data.close()

        return ret, resp

    def __upload_data_with_retrier(
        self,
        access_key,
        bucket_name,
        **upload_data_opts
    ):
        retrier = get_default_retrier(
            regions_provider=self._get_regions_provider(
                access_key=access_key,
                bucket_name=bucket_name
            ),
            accelerate_uploading=self.accelerate_uploading
        )
        data = upload_data_opts.get('data')
        attempt = None
        for attempt in retrier:
            with attempt:
                attempt.result = self.__upload_data(
                    up_endpoint=attempt.context.get('endpoint'),
                    **upload_data_opts
                )
                ret, resp = attempt.result
                if resp.ok() and ret:
                    return attempt.result
                if (
                    not is_seekable(data) or
                    not resp.need_retry()
                ):
                    return attempt.result
                data.seek(0)

        if attempt is None:
            raise RuntimeError('Retrier is not working. attempt is None')

        return attempt.result

    def __upload_data(
        self,
        up_endpoint,
        fields,
        file_name,
        data,
        data_size=None,
        mime_type='application/octet-stream'
    ):
        """
        Parameters
        ----------
        up_endpoint: Endpoint
        fields: dict
        file_name: str
        data: IOBase
        data_size: int
        mime_type: str

        Returns
        -------
        ret: dict
        resp: ResponseInfo
        """
        req_url = up_endpoint.get_value(scheme=self.preferred_scheme)
        if not file_name or not file_name.strip():
            file_name = 'file_name'

        ret, resp = qn_http_client.post(
            url=req_url,
            data=fields,
            files={
                'file': (file_name, data, mime_type)
            }
        )
        return ret, resp

    def __get_form_fields(
        self,
        up_token,
        **kwargs
    ):
        """
        Parameters
        ----------
        up_token: str
        kwargs
            key, crc32_int, custom_vars, metadata

        Returns
        -------
        dict
        """
        key = kwargs.get('key', None)
        crc32_int = kwargs.get('crc32_int', None)
        custom_vars = kwargs.get('custom_vars', None)
        metadata = kwargs.get('metadata', None)

        result = {
            'token': up_token,
        }

        if key is not None:
            result['key'] = key

        if crc32_int:
            result['crc32'] = crc32_int

        if custom_vars:
            result.update(
                {
                    k: str(v)
                    for k, v in custom_vars.items()
                    if k.startswith('x:')
                }
            )

        if metadata:
            result.update(
                {
                    k: str(v)
                    for k, v in metadata.items()
                    if k.startswith('x-qn-meta-')
                }
            )

        return result

    def __get_crc32_int(self, data):
        """
        Parameters
        ----------
        data: BytesIO

        Returns
        -------
        str
        """
        result = None
        if not is_seekable(data):
            return result
        result = io_crc32(data)
        data.seek(0)
        return result



# -*- coding: utf-8 -*-
from functools import reduce


def compose_middleware(middlewares, handle):
    """
    Args:
        middlewares (list[Middleware]): Middlewares
        handle ((requests.Request) -> qiniu.http.response.ResponseInfo): The send request handle

    Returns:
        (requests.Request) -> qiniu.http.response.ResponseInfo: Composed handle

    """
    middlewares.reverse()

    return reduce(
        lambda h, mw:
            lambda req: mw(req, h),
        middlewares,
        handle
    )


class Middleware:
    def __call__(self, request, nxt):
        """
        Args:
            request (requests.Request):
            nxt ((requests.Request) -> qiniu.http.response.ResponseInfo):

        Returns:
            requests.Response:

        """
        raise NotImplementedError('{0}.__call__ method is not implemented yet'.format(type(self)))



# -*- coding: utf-8 -*-
import base64

from .commands import CommandHandle, command_result_from_events
from .envd import connect_rpc, connect_stream_rpc
from .errors import SandboxError


class PtySize(object):
    def __init__(self, rows=24, cols=80):
        self.rows = rows
        self.cols = cols

    def to_dict(self):
        return {'rows': self.rows, 'cols': self.cols}


def _normalize_size(size=None, rows=None, cols=None):
    if size is None:
        return {'rows': rows or 24, 'cols': cols or 80}
    if hasattr(size, 'to_dict'):
        data = size.to_dict()
    elif isinstance(size, dict):
        data = size
    else:
        data = {
            'rows': getattr(size, 'rows', None),
            'cols': getattr(size, 'cols', None),
        }
    return {
        'rows': data.get('rows') or rows or 24,
        'cols': data.get('cols') or cols or 80,
    }


class Pty(object):
    def __init__(self, sandbox):
        self.sandbox = sandbox

    def create(self, size=None, user=None, cwd=None, envs=None, timeout=None,
               rows=None, cols=None, **opts):
        envs = dict(envs or {})
        envs.setdefault('TERM', 'xterm-256color')
        envs.setdefault('LANG', 'C.UTF-8')
        envs.setdefault('LC_ALL', 'C.UTF-8')
        process = {
            'cmd': '/bin/bash',
            'args': ['-i', '-l'],
            'envs': envs,
        }
        if cwd:
            process['cwd'] = cwd
        body = {
            'process': process,
            'pty': {'size': _normalize_size(size, rows=rows, cols=cols)},
        }
        if opts.get('tag'):
            body['tag'] = opts.get('tag')
        events = connect_stream_rpc(
            self.sandbox,
            '/process.Process/Start',
            body,
            user=user,
            timeout=timeout,
            stream=True,
        )
        events = iter(events)
        try:
            first_event = next(events)
        except StopIteration:
            first_event = None
        result = command_result_from_events([first_event])
        return CommandHandle(self, result, events=events)

    def connect(self, pid, user=None, timeout=None, throw_on_error=False):
        events = connect_stream_rpc(
            self.sandbox,
            '/process.Process/Connect',
            {'process': {'selector': {'pid': pid}}},
            user=user,
            timeout=timeout,
            stream=True,
        )
        events = iter(events)
        try:
            first_event = next(events)
        except StopIteration:
            first_event = None
        result = command_result_from_events([first_event])
        return CommandHandle(
            self,
            result,
            events=events,
            throw_on_error=throw_on_error,
        )

    def send_stdin(self, pid, data, user=None, timeout=None):
        if isinstance(data, bytearray):
            data = bytes(data)
        elif not isinstance(data, bytes):
            if hasattr(data, 'encode'):
                data = data.encode('utf-8')
            else:
                data = str(data).encode('utf-8')
        return connect_rpc(self.sandbox, '/process.Process/SendInput', {
            'process': {'selector': {'pid': pid}},
            'input': {'pty': base64.b64encode(data).decode('ascii')},
        }, user=user, timeout=timeout)

    sendStdin = send_stdin
    send_input = send_stdin
    sendInput = send_stdin

    def resize(self, pid, size=None, user=None, timeout=None, rows=None,
               cols=None):
        return connect_rpc(self.sandbox, '/process.Process/Update', {
            'process': {'selector': {'pid': pid}},
            'pty': {'size': _normalize_size(size, rows=rows, cols=cols)},
        }, user=user, timeout=timeout)

    def kill(self, pid, user=None, timeout=None):
        try:
            connect_rpc(self.sandbox, '/process.Process/SendSignal', {
                'process': {'selector': {'pid': pid}},
                'signal': 'SIGNAL_SIGKILL',
            }, user=user, timeout=timeout)
            return True
        except SandboxError as err:
            if err.status_code == 404:
                return False
            raise



# -*- coding: utf-8 -*-
import json
import struct

import requests

from .constants import DEFAULT_USER
from .errors import SandboxError
from .util import basic_auth


MAX_CONNECT_ENVELOPE_BYTES = 10 * 1024 * 1024


def envd_headers(sandbox, user=None, extra=None):
    headers = {
        'Authorization': basic_auth(user or DEFAULT_USER),
    }
    if sandbox.envd_access_token:
        headers['X-Access-Token'] = sandbox.envd_access_token
    if extra:
        headers.update(extra)
    return headers


def connect_rpc(sandbox, procedure, body=None, user=None, timeout=None):
    if timeout is None:
        timeout = sandbox.client.timeout
    url = sandbox.envd_url() + procedure
    headers = envd_headers(sandbox, user, {'Content-Type': 'application/json'})
    try:
        response = sandbox.client.session.post(
            url,
            data=json.dumps(body or {}, separators=(',', ':')),
            headers=headers,
            timeout=timeout,
        )
    except requests.RequestException as err:
        raise SandboxError('Sandbox envd request failed: {0}'.format(err))
    if response.status_code < 200 or response.status_code >= 300:
        raise SandboxError(
            'Sandbox envd request failed with status {0}'.format(
                response.status_code), response, response.text, )
    if not response.content:
        return None
    data = response.json()
    if isinstance(data, dict) and 'result' in data:
        return data.get('result')
    return data


def encode_connect_envelope(message):
    payload = json.dumps(message or {}, separators=(',', ':')).encode('utf-8')
    return struct.pack('>BI', 0, len(payload)) + payload


def decode_connect_envelopes(data):
    if not data:
        return []
    if not isinstance(data, bytes):
        data = data.encode('utf-8')
    messages = []
    offset = 0
    while offset + 5 <= len(data):
        flags, length = struct.unpack('>BI', data[offset:offset + 5])
        offset += 5
        if length > MAX_CONNECT_ENVELOPE_BYTES:
            raise SandboxError(
                'Sandbox envd stream envelope too large: {0}'.format(length)
            )
        if offset + length > len(data):
            raise SandboxError('Sandbox envd stream truncated unexpectedly')
        payload = data[offset:offset + length]
        offset += length
        if flags & 2:
            if _is_connect_end(payload):
                continue
            _raise_connect_error(payload)
        if payload:
            messages.append(json.loads(payload.decode('utf-8')))
    if offset < len(data):
        raise SandboxError('Sandbox envd stream truncated unexpectedly')
    return messages


def _raise_connect_error(payload):
    error = None
    message = 'Sandbox envd stream failed'
    if payload:
        try:
            data = json.loads(payload.decode('utf-8'))
            if isinstance(data, dict):
                error = data.get('error')
                if isinstance(error, dict) and error.get('message'):
                    message = error.get('message')
        except (TypeError, ValueError):
            pass
    raise SandboxError(message, data=error)


def _is_connect_end(payload):
    if not payload:
        return False
    try:
        data = json.loads(payload.decode('utf-8'))
    except (TypeError, ValueError):
        return False
    return data == {}


def iter_connect_envelopes(chunks, response=None):
    try:
        buffer = b''
        for chunk in chunks:
            if not chunk:
                continue
            if not isinstance(chunk, bytes):
                chunk = chunk.encode('utf-8')
            buffer += chunk
            while len(buffer) >= 5:
                flags, length = struct.unpack('>BI', buffer[:5])
                if length > MAX_CONNECT_ENVELOPE_BYTES:
                    raise SandboxError(
                        'Sandbox envd stream envelope too large: {0}'.format(
                            length)
                    )
                if len(buffer) < 5 + length:
                    break
                payload = buffer[5:5 + length]
                buffer = buffer[5 + length:]
                if flags & 2:
                    if _is_connect_end(payload):
                        continue
                    _raise_connect_error(payload)
                if payload:
                    yield json.loads(payload.decode('utf-8'))
        if buffer:
            raise SandboxError('Sandbox envd stream truncated unexpectedly')
    finally:
        if response is not None:
            try:
                response.close()
            except Exception:
                pass


def connect_stream_rpc(sandbox, procedure, body=None, user=None, timeout=None,
                       stream=False):
    if timeout is None:
        timeout = sandbox.client.timeout
    url = sandbox.envd_url() + procedure
    headers = envd_headers(sandbox, user, {
        'Content-Type': 'application/connect+json',
        'Keepalive-Ping-Interval': '50',
    })
    request_opts = {
        'data': encode_connect_envelope(body),
        'headers': headers,
        'timeout': timeout,
    }
    if stream:
        request_opts['stream'] = True
    try:
        try:
            response = sandbox.client.session.post(url, **request_opts)
        except TypeError:
            request_opts.pop('stream', None)
            response = sandbox.client.session.post(url, **request_opts)
    except requests.RequestException as err:
        raise SandboxError('Sandbox envd request failed: {0}'.format(err))
    if response.status_code < 200 or response.status_code >= 300:
        raise SandboxError(
            'Sandbox envd request failed with status {0}'.format(
                response.status_code), response, response.text, )

    if stream and hasattr(response, 'iter_content'):
        return iter_connect_envelopes(
            response.iter_content(chunk_size=8192),
            response=response,
        )

    content_type = response.headers.get('Content-Type', '')
    if 'application/connect+json' in content_type:
        return decode_connect_envelopes(response.content)

    if not response.content:
        return []
    data = response.json()
    if isinstance(data, dict) and 'result' in data:
        data = data.get('result')
    if isinstance(data, dict) and 'events' in data:
        return data.get('events')
    if isinstance(data, dict) and 'event' in data:
        return [data]
    return data


def raw_envd_request(sandbox, method, url, **kwargs):
    if kwargs.get('timeout') is None:
        kwargs['timeout'] = sandbox.client.timeout
    try:
        response = sandbox.client.session.request(method, url, **kwargs)
    except requests.RequestException as err:
        raise SandboxError('Sandbox envd request failed: {0}'.format(err))
    if response.status_code < 200 or response.status_code >= 300:
        raise SandboxError(
            'Sandbox envd request failed with status {0}'.format(
                response.status_code), response, response.text, )
    return response



# -*- coding: utf-8 -*-
import functools
import logging
import os
import time


from .compat import json, s as str_from_bytes
from .utils import urlsafe_base64_decode, dt2ts
from .config import UC_HOST, is_customized_default, get_default
from .http.endpoint import Endpoint as _HTTPEndpoint
from .http.regions_provider import Region as _HTTPRegion, ServiceName, get_default_regions_provider


def _legacy_default_get(key):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            if hasattr(self, key) and getattr(self, key):
                return getattr(self, key)
            if is_customized_default('default_' + key):
                return get_default('default_' + key)
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class LegacyRegion(_HTTPRegion, object):
    """七牛上传区域类
    该类主要内容上传区域地址。
    """

    def __init__(
        self,
        up_host=None,
        up_host_backup=None,
        io_host=None,
        host_cache=None,
        home_dir=None,
        scheme="http",
        rs_host=None,
        rsf_host=None,
        api_host=None,
        accelerate_uploading=False
    ):
        """初始化Zone类"""
        super(LegacyRegion, self).__init__()
        if host_cache is None:
            host_cache = {}
        self.up_host = up_host
        self.up_host_backup = up_host_backup
        self.io_host = io_host
        self.rs_host = rs_host
        self.rsf_host = rsf_host
        self.api_host = api_host
        self.home_dir = home_dir
        self.host_cache = host_cache
        self.scheme = scheme
        self.services.update({
            k: [
                _HTTPEndpoint.from_host(h)
                for h in v if h
            ]
            for k, v in {
                ServiceName.UP: [up_host, up_host_backup],
                ServiceName.IO: [io_host],
                ServiceName.RS: [rs_host],
                ServiceName.RSF: [rsf_host],
                ServiceName.API: [api_host]
            }.items()
        })
        self.accelerate_uploading = accelerate_uploading

    def get_up_host_by_token(self, up_token, home_dir):
        ak, bucket = self.unmarshal_up_token(up_token)
        if home_dir is None:
            home_dir = os.getcwd()
        up_hosts = self.get_up_host(ak, bucket, home_dir)
        return up_hosts[0]

    def get_up_host_backup_by_token(self, up_token, home_dir):
        ak, bucket = self.unmarshal_up_token(up_token)
        if home_dir is None:
            home_dir = os.getcwd()
        up_hosts = self.get_up_host(ak, bucket, home_dir)
        if len(up_hosts) <= 1:
            up_host = up_hosts[0]
        else:
            up_host = up_hosts[1]
        return up_host

    def get_io_host(self, ak, bucket, home_dir=None):
        if self.io_host:
            return self.io_host
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'ioHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        io_hosts = bucket_hosts['ioHosts']
        return io_hosts[0]

    @_legacy_default_get('rs_host')
    def get_rs_host(self, ak, bucket, home_dir=None):
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'rsHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        rs_hosts = bucket_hosts['rsHosts']
        return rs_hosts[0]

    @_legacy_default_get('rsf_host')
    def get_rsf_host(self, ak, bucket, home_dir=None):
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'rsfHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        rsf_hosts = bucket_hosts['rsfHosts']
        return rsf_hosts[0]

    @_legacy_default_get('api_host')
    def get_api_host(self, ak, bucket, home_dir=None):
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'apiHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        api_hosts = bucket_hosts['apiHosts']
        return api_hosts[0]

    def get_up_host(self, ak, bucket, home_dir):
        if home_dir is None:
            home_dir = os.getcwd()
        bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir)
        if 'upHosts' not in bucket_hosts:
            bucket_hosts = self.get_bucket_hosts(ak, bucket, home_dir, force=True)
        up_hosts = bucket_hosts['upHosts']
        return up_hosts

    def unmarshal_up_token(self, up_token):
        token = up_token.split(':')
        if len(token) != 3:
            raise ValueError('invalid up_token')

        ak = token[0]
        policy = json.loads(
            str_from_bytes(
                urlsafe_base64_decode(
                    token[2])))

        scope = policy["scope"]
        bucket = scope
        if ':' in scope:
            bucket = scope.split(':')[0]

        return ak, bucket

    def get_bucket_hosts(self, ak, bucket, home_dir=None, force=False):
        cache_persist_path = os.path.join(home_dir, 'qn-regions-cache.jsonl') if home_dir else None
        regions = self.__get_bucket_regions(
            ak,
            bucket,
            force_query=force,
            cache_persist_path=cache_persist_path
        )

        if not regions:
            raise KeyError("Please check your BUCKET_NAME! Server hosts not correct! The hosts is empty")

        region = regions[0]

        bucket_hosts = {
            k: [
                e.get_value(scheme=self.scheme)
                for e in region.services[sn]
                if e
            ]
            for k, sn in {
                'upHosts': ServiceName.UP,
                'ioHosts': ServiceName.IO,
                'rsHosts': ServiceName.RS,
                'rsfHosts': ServiceName.RSF,
                'apiHosts': ServiceName.API
            }.items()
        }

        ttl = region.ttl if region.ttl > 0 else 24 * 3600  # 1 day
        # use datetime.datetime.timestamp() when min version of python >= 3
        create_time = dt2ts(region.create_time)
        bucket_hosts['deadline'] = create_time + ttl

        return bucket_hosts

    def get_bucket_hosts_to_cache(self, key, home_dir):
        """
        .. deprecated::
            The cache has been replaced by CachedRegionsProvider

        Parameters
        ----------
        key: str
        home_dir: str

        Returns
        -------
        dict
        """
        ret = {}
        if len(self.host_cache) == 0:
            self.host_cache_from_file(home_dir)

        if key not in self.host_cache:
            return ret

        if self.host_cache[key]['deadline'] > time.time():
            ret = self.host_cache[key]

        return ret

    def set_bucket_hosts_to_cache(self, key, val, home_dir):
        """
        .. deprecated::
            The cache has been replaced by CachedRegionsProvider

        Parameters
        ----------
        key: str
        val: dict
        home_dir: str
        """
        self.host_cache[key] = val
        self.host_cache_to_file(home_dir)
        return

    def host_cache_from_file(self, home_dir):
        """
        .. deprecated::
            The cache has been replaced by CachedRegionsProvider

        Parameters
        ----------
        home_dir: str
        """
        if home_dir is not None:
            self.home_dir = home_dir
        path = self.host_cache_file_path()
        if not os.path.isfile(path):
            return None
        with open(path, 'r') as f:
            try:
                bucket_hosts = json.load(f)
                self.host_cache = bucket_hosts
            except Exception as e:
                logging.error(e)
        f.close()
        return

    def host_cache_file_path(self):
        """
        .. deprecated::
            The cache has been replaced by CachedRegionsProvider

        Returns
        -------
        str
        """
        return os.path.join(self.home_dir, ".qiniu_pythonsdk_hostscache.json")

    def host_cache_to_file(self, home_dir):
        """
        .. deprecated::
            The cache has been replaced by CachedRegionsProvider

        Parameters
        ----------
        home_dir: str

        """
        path = self.host_cache_file_path()
        with open(path, 'w') as f:
            json.dump(self.host_cache, f)
        f.close()

    def bucket_hosts(self, ak, bucket):
        regions = self.__get_bucket_regions(ak, bucket)

        data_dict = {
            'hosts': [
                {
                    k.value if isinstance(k, ServiceName) else k: {
                        'domains': [
                            e.host for e in v
                        ]
                    }
                    for k, v in r.services.items()
                }
                for r in regions
            ]
        }
        for r in data_dict['hosts']:
            if 'up_acc' in r:
                r.setdefault('up', {})
                r['up'].update(acc_domains=r['up_acc'].get('domains', []))
                del r['up_acc']

        data = json.dumps(data_dict)

        return data

    def __get_bucket_regions(
        self,
        access_key,
        bucket_name,
        force_query=False,
        cache_persist_path=None
    ):
        query_region_host = UC_HOST
        if is_customized_default('default_query_region_host'):
            query_region_host = get_default('default_query_region_host')
        query_region_backup_hosts = get_default('default_query_region_backup_hosts')
        query_region_backup_retry_times = get_default('default_backup_hosts_retry_times')

        regions_provider = get_default_regions_provider(
            query_endpoints_provider=[
                _HTTPEndpoint.from_host(h)
                for h in [query_region_host] + query_region_backup_hosts
                if h
            ],
            access_key=access_key,
            bucket_name=bucket_name,
            accelerate_uploading=self.accelerate_uploading,
            force_query=force_query,
            preferred_scheme=self.scheme,
            persist_path=cache_persist_path,
            max_retry_times_per_endpoint=query_region_backup_retry_times
        )

        return list(regions_provider)


Region = LegacyRegion



# -*- coding: utf-8 -*-

from qiniu import config
from qiniu import http


class PersistentFop(object):
    """持久化处理类

    该类用于主动触发异步持久化操作，具体规格参考：
    https://developer.qiniu.com/dora/api/persistent-data-processing-pfop

    Attributes:
        auth:       账号管理密钥对，Auth对象
        bucket:     操作资源所在空间
        pipeline:   多媒体处理队列，详见 https://developer.qiniu.com/dora/6499/tasks-and-workflows
        notify_url: 持久化处理结果通知URL
    """

    def __init__(self, auth, bucket, pipeline=None, notify_url=None):
        """初始化持久化处理类"""
        self.auth = auth
        self.bucket = bucket
        self.pipeline = pipeline
        self.notify_url = notify_url

    def execute(self, key, fops=None, force=None, persistent_type=None, workflow_template_id=None):
        """
        执行持久化处理

        Parameters
        ----------
        key: str
            待处理的源文件
        fops: list[str], optional
            处理详细操作，规格详见 https://developer.qiniu.com/dora/manual/1291/persistent-data-processing-pfop
            与 template_id 二选一
        force: int or str, optional
            强制执行持久化处理开关
        persistent_type: int or str, optional
            持久化处理类型，为 '1' 时开启闲时任务
        template_id: str, optional
            与 fops 二选一
        Returns
        -------
        ret: dict
            持久化处理的 persistentId，类似 {"persistentId": 5476bedf7823de4068253bae};
        resp: ResponseInfo
        """
        if not fops and not workflow_template_id:
            raise ValueError('Must provide one of fops or template_id')
        data = {
            'bucket': self.bucket,
            'key': key,
        }
        if self.pipeline:
            data['pipeline'] = self.pipeline
        if self.notify_url:
            data['notifyURL'] = self.notify_url
        if fops:
            data['fops'] = ';'.join(fops)
        if force == 1 or force == '1':
            data['force'] = str(force)
        if persistent_type and type(int(persistent_type)) is int:
            data['type'] = str(persistent_type)
        if workflow_template_id:
            data['workflowTemplateID'] = workflow_template_id

        url = '{0}/pfop'.format(config.get_default('default_api_host'))
        return http._post_with_auth(url, data, self.auth)

    def get_status(self, persistent_id):
        """
        获取持久化处理状态

        Parameters
        ----------
        persistent_id: str

        Returns
        -------
        ret: dict
            持久化处理的状态，详见 https://developer.qiniu.com/dora/1294/persistent-processing-status-query-prefop
        resp: ResponseInfo
        """
        url = '{0}/status/get/prefop'.format(config.get_default('default_api_host'))
        data = {
            'id': persistent_id
        }
        return http._get_with_auth(url, data, self.auth)



import functools

from .attempt import Attempt


def before_retry_nothing(attempt, policy):
    return True


class Retrier:
    def __init__(self, policies=None, before_retry=None):
        """
        Parameters
        ----------
        policies: list[qiniu.retry.abc.RetryPolicy]
        before_retry: callable
            `(attempt: Attempt, policy: qiniu.retry.abc.RetryPolicy) -> bool`
        """
        self.policies = policies if policies is not None else []
        self.before_retry = before_retry if before_retry is not None else before_retry_nothing

    def __iter__(self):
        retrying = Retrying(
            # change to `list.copy` for more readable when min version of python update to >= 3
            policies=self.policies[:],
            before_retry=self.before_retry
        )
        retrying.init_context()
        while True:
            attempt = Attempt(retrying.context)
            yield attempt
            if (
                hasattr(attempt.exception, 'no_need_retry') and
                attempt.exception.no_need_retry
            ):
                break
            policy = retrying.get_retry_policy(attempt)
            if not policy:
                break
            if not self.before_retry(attempt, policy):
                break
            policy.prepare_retry(attempt)
            retrying.after_retried(attempt, policy)
        if attempt.exception:
            raise attempt.exception

    def try_do(
        self,
        func,
        *args,
        **kwargs
    ):
        attempt = None
        for attempt in self:
            with attempt:
                if kwargs.get('with_retry_context', False):
                    # inject retry_context
                    kwargs['retry_context'] = attempt.context
                if 'with_retry_context' in kwargs:
                    del kwargs['with_retry_context']

                # store result
                attempt.result = func(*args, **kwargs)

        if attempt is None:
            raise RuntimeError('attempt is none')

        return attempt.result

    def _wrap(self, with_retry_context=False):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return self.try_do(
                    func,
                    with_retry_context=with_retry_context,
                    *args,
                    **kwargs
                )

            return wrapper

        return decorator

    def retry(self, *args, **kwargs):
        """
        decorator to retry
        """
        if len(args) == 1 and callable(args[0]):
            return self.retry()(args[0])
        else:
            return self._wrap(**kwargs)


class Retrying:
    def __init__(self, policies, before_retry):
        """
        Parameters
        ----------
        policies: list[qiniu.retry.abc.RetryPolicy]
        before_retry: callable
            `(attempt: Attempt, policy: qiniu.retry.abc.RetryPolicy) -> bool`
        """
        self.policies = policies
        self.before_retry = before_retry
        self.context = {}

    def init_context(self):
        for policy in self.policies:
            policy.init_context(self.context)

    def get_retry_policy(self, attempt):
        """

        Parameters
        ----------
        attempt: Attempt

        Returns
        -------
        qiniu.retry.abc.RetryPolicy

        """
        policy = None

        # find important policy
        for p in self.policies:
            if p.is_important(attempt):
                policy = p
                break
        if policy and policy.should_retry(attempt):
            return policy
        else:
            policy = None

        # find retry policy
        for p in self.policies:
            if p.should_retry(attempt):
                policy = p
                break

        return policy

    def after_retried(self, attempt, policy):
        for p in self.policies:
            p.after_retry(attempt, policy)


"""
Examples
--------
retrier = Retrier()
result = None
for attempt in retrier:
    with attempt:
        endpoint = attempt.context.get('endpoint')
        result = upload(endpoint)
        attempt.result = result
return result
"""

"""
Examples
--------
def foo():
    print('hi')

retrier = Retrier()
retrier.try_do(foo)
"""

"""
Examples
--------
retrier = Retrier()


@retrier.retry
def foo():
    print('hi')

foo()
"""



class Attempt:
    def __init__(self, custom_context=None):
        """
        Parameters
        ----------
        custom_context: dict or None
        """
        self.context = custom_context if custom_context is not None else {}
        self.exception = None
        self.result = None

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None and exc_val is not None:
            self.exception = exc_val
            return True  # Swallow exception.



from qiniu.compat import urlparse

from .base import Middleware


class RetryDomainsMiddleware(Middleware):
    def __init__(self, backup_domains, max_retry_times=2, retry_condition=None):
        """
        Args:
            backup_domains (list[str]):
            max_retry_times (int):
            retry_condition ((requests.Response or None, requests.Request)->bool):
        """
        self.backup_domains = backup_domains
        self.max_retry_times = max_retry_times
        self.retry_condition = retry_condition

        self.retried_times = 0

    @staticmethod
    def _get_changed_url(url, domain):
        url_parse_result = urlparse(url)

        backup_netloc = ''
        has_user = False
        if url_parse_result.username is not None:
            backup_netloc += url_parse_result.username
            has_user = True
        if url_parse_result.password is not None:
            backup_netloc += url_parse_result.password
            has_user = True
        if has_user:
            backup_netloc += '@'
        backup_netloc += domain
        if url_parse_result.port is not None:
            backup_netloc += ':' + str(url_parse_result.port)

        # the _replace is a public method. start with `_` just to prevent conflicts with field names
        # see namedtuple docs
        url_parse_result = url_parse_result._replace(
            netloc=backup_netloc
        )

        return url_parse_result.geturl()

    @staticmethod
    def _try_nxt(request, nxt):
        resp = None
        err = None
        try:
            resp = nxt(request)
        except Exception as e:
            err = e
        return resp, err

    def _should_retry(self, resp, req):
        if callable(self.retry_condition):
            return self.retry_condition(resp, req)

        return resp is None or resp.need_retry()

    def __call__(self, request, nxt):
        resp_info, err = None, None
        url_parse_result = urlparse(request.url)

        for backup_domain in [str(url_parse_result.hostname)] + self.backup_domains:
            request.url = RetryDomainsMiddleware._get_changed_url(request.url, backup_domain)
            self.retried_times = 0

            while self.retried_times < self.max_retry_times:
                resp_info, err = RetryDomainsMiddleware._try_nxt(request, nxt)
                self.retried_times += 1
                if not self._should_retry(resp_info, request):
                    if err is not None:
                        raise err
                    return resp_info

        if err is not None:
            raise err

        return resp_info



# -*- coding: utf-8 -*-
from io import IOBase, TextIOBase

from qiniu.compat import basestring

from .errors import InvalidArgumentException, SandboxError
from .envd import connect_rpc, envd_headers, raw_envd_request
from .util import file_basename


class _EncodedTextReader(object):
    def __init__(self, stream, encoding):
        self.stream = stream
        self.encoding = encoding

    def read(self, size=-1):
        chunk = self.stream.read(size)
        if isinstance(chunk, basestring):
            return chunk.encode(self.encoding)
        return chunk

    def __getattr__(self, name):
        if name in ('seek', 'tell', 'getvalue', 'len'):
            raise AttributeError(name)
        return getattr(self.stream, name)


class FileType(object):
    FILE = 'file'
    DIR = 'dir'
    DIRECTORY = DIR


class FilesystemEventType(object):
    CREATE = 'create'
    WRITE = 'write'
    REMOVE = 'remove'
    RENAME = 'rename'
    CHMOD = 'chmod'


class FilesystemEvent(object):
    def __init__(self, name=None, type=None):
        self.name = name
        self.type = type

    def to_dict(self):
        return {'name': self.name, 'type': self.type}


class WriteEntry(dict):
    def __init__(self, path=None, data=None, **kwargs):
        dict.__init__(self, path=path, data=data, **kwargs)
        self.path = path
        self.data = data


class WriteInfo(object):
    def __init__(self, name=None, type=None, path=None):
        self.name = name
        self.type = type
        self.path = path

    def to_dict(self):
        return {
            'name': self.name,
            'type': self.type,
            'path': self.path,
        }

    def __getitem__(self, key):
        return self.to_dict()[key]

    def __eq__(self, other):
        if isinstance(other, dict):
            data = self.to_dict()
            return all(data.get(key) == value for key, value in other.items())
        return object.__eq__(self, other)


class EntryInfo(WriteInfo):
    def __init__(self, name=None, type=None, path=None, size=None, mode=None,
                 permissions=None, owner=None, group=None, modified_time=None,
                 symlink_target=None):
        WriteInfo.__init__(self, name=name, type=type, path=path)
        self.size = size
        self.mode = mode
        self.permissions = permissions
        self.owner = owner
        self.group = group
        self.modified_time = modified_time
        self.symlink_target = symlink_target

    def to_dict(self):
        data = WriteInfo.to_dict(self)
        data.update({
            'size': self.size,
            'mode': self.mode,
            'permissions': self.permissions,
            'owner': self.owner,
            'group': self.group,
            'modified_time': self.modified_time,
            'symlink_target': self.symlink_target,
        })
        return data


def normalize_event_type(event_type):
    mapping = {
        'EVENT_TYPE_CREATE': FilesystemEventType.CREATE,
        'EVENT_TYPE_WRITE': FilesystemEventType.WRITE,
        'EVENT_TYPE_REMOVE': FilesystemEventType.REMOVE,
        'EVENT_TYPE_RENAME': FilesystemEventType.RENAME,
        'EVENT_TYPE_CHMOD': FilesystemEventType.CHMOD,
        1: FilesystemEventType.CREATE,
        2: FilesystemEventType.WRITE,
        3: FilesystemEventType.REMOVE,
        4: FilesystemEventType.RENAME,
        5: FilesystemEventType.CHMOD,
    }
    return mapping.get(event_type, event_type)


def normalize_entry_type(entry_type):
    if entry_type in ('FILE_TYPE_DIRECTORY', 'DIRECTORY', 'dir'):
        return FileType.DIR
    if entry_type in ('FILE_TYPE_FILE', 'FILE', 'file'):
        return FileType.FILE
    return entry_type


def normalize_entry(entry, extended=False):
    entry = entry or {}
    entry_type = entry.get('type')
    entry_type = normalize_entry_type(entry_type)
    info_cls = EntryInfo if extended else WriteInfo
    return info_cls(
        name=entry.get('name'),
        type=entry_type,
        path=entry.get('path'),
        size=entry.get('size'),
        mode=entry.get('mode'),
        permissions=entry.get('permissions'),
        owner=entry.get('owner'),
        group=entry.get('group'),
        modified_time=entry.get('modifiedTime') or entry.get('modified_time'),
        symlink_target=entry.get('symlinkTarget') or entry.get(
            'symlink_target'),
    ) if extended else info_cls(
        name=entry.get('name'),
        type=entry_type,
        path=entry.get('path'),
    )


def to_upload_body(data, encoding='utf-8'):
    if isinstance(data, bytearray):
        return bytes(data)
    if isinstance(data, bytes):
        return data
    if isinstance(data, basestring):
        return data.encode(encoding)
    if isinstance(data, TextIOBase):
        return _EncodedTextReader(data, encoding)
    if isinstance(data, IOBase) or hasattr(data, 'read'):
        return data
    raise InvalidArgumentException(
        'Unsupported data type for filesystem write: {0}'.format(type(data)))


def _response_stream(response, chunk_size=8192):
    try:
        for chunk in response.iter_content(chunk_size=chunk_size):
            yield chunk
    finally:
        response.close()


class WatchHandle(object):
    def __init__(self, filesystem, watcher_id):
        self.filesystem = filesystem
        self.watcher_id = watcher_id
        self.watcherID = watcher_id
        self._closed = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False

    def get_new_events(self, user=None, timeout=None):
        if self._closed:
            raise SandboxError('The watcher is already stopped')
        data = connect_rpc(
            self.filesystem.sandbox,
            '/filesystem.Filesystem/GetWatcherEvents',
            {'watcherId': self.watcher_id},
            user=user,
            timeout=timeout,
        )
        events = []
        for event in (data or {}).get('events', []):
            events.append(FilesystemEvent(
                name=event.get('name'),
                type=normalize_event_type(event.get('type')),
            ))
        return events

    getNewEvents = get_new_events

    def stop(self, user=None, timeout=None):
        if self._closed:
            return None
        connect_rpc(
            self.filesystem.sandbox,
            '/filesystem.Filesystem/RemoveWatcher',
            {'watcherId': self.watcher_id},
            user=user,
            timeout=timeout,
        )
        self._closed = True
        return None


class Filesystem(object):
    def __init__(self, sandbox):
        self.sandbox = sandbox

    def _request_timeout(self, opts):
        if opts.get('request_timeout') is not None:
            return opts.get('request_timeout')
        return opts.get('timeout')

    def read(self, path, user=None, fmt=None, **opts):
        if fmt is None:
            fmt = opts.pop('format', 'text')
        url = self.sandbox.download_url(path, user=user)
        stream_mode = fmt == 'stream'
        response = raw_envd_request(
            self.sandbox,
            'GET',
            url,
            headers=envd_headers(self.sandbox, user),
            timeout=self._request_timeout(opts),
            stream=stream_mode,
        )
        if fmt == 'stream':
            if hasattr(response, 'iter_content'):
                return _response_stream(
                    response,
                    chunk_size=opts.get('chunk_size', 8192),
                )
            return iter([response.content])
        if fmt == 'bytes':
            return bytearray(response.content)
        return response.content.decode(opts.get('encoding', 'utf-8'))

    def read_text(self, path, user=None, **opts):
        return self.read(path, user=user, format='text', **opts)

    readText = read_text

    def write(self, path, data, user=None, **opts):
        data = to_upload_body(data, opts.get('encoding', 'utf-8'))
        url = self.sandbox.upload_url(path, user=user)
        if opts.get('use_octet_stream'):
            response = raw_envd_request(
                self.sandbox,
                'POST',
                url,
                data=data,
                headers=envd_headers(
                    self.sandbox,
                    user,
                    {'Content-Type': 'application/octet-stream'},
                ),
                timeout=self._request_timeout(opts),
            )
            return self._format_write_response(response)

        response = raw_envd_request(
            self.sandbox,
            'POST',
            url,
            files={'file': (file_basename(path), data)},
            headers=envd_headers(self.sandbox, user),
            timeout=self._request_timeout(opts),
        )
        return self._format_write_response(response)

    def _format_write_response(self, response):
        if not response.content:
            return None
        data = response.json()
        if isinstance(data, list):
            return normalize_entry(data[0] if data else {})
        return normalize_entry(data)

    def write_files(self, files, user=None, **opts):
        result = []
        for item in files or []:
            if isinstance(item, dict):
                path = item.get('path')
                data = item.get('data', item.get('content', ''))
            else:
                path, data = item
            result.append(self.write(path, data, user=user, **opts))
        return result

    writeFiles = write_files

    def get_info(self, path, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/Stat',
            {'path': path},
            user=user,
            timeout=timeout,
        )
        return normalize_entry((data or {}).get('entry'), extended=True)

    getInfo = get_info
    stat = get_info

    def list(self, path, depth=1, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/ListDir',
            {'path': path, 'depth': depth},
            user=user,
            timeout=timeout,
        )
        return [
            normalize_entry(entry, extended=True) for entry in (
                data or {}).get(
                'entries', [])]

    def exists(self, path, user=None, timeout=None):
        try:
            self.get_info(path, user=user, timeout=timeout)
            return True
        except SandboxError as err:
            response = getattr(err, 'response', None)
            if response is not None and getattr(
                    response, 'status_code', None) == 404:
                return False
            raise

    def make_dir(self, path, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/MakeDir',
            {'path': path},
            user=user,
            timeout=timeout,
        )
        return normalize_entry((data or {}).get('entry'), extended=True)

    makeDir = make_dir
    mkdir = make_dir

    def remove(self, path, user=None, timeout=None):
        connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/Remove',
            {'path': path},
            user=user,
            timeout=timeout,
        )
        return None

    def rename(self, old_path, new_path, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/Move',
            {'source': old_path, 'destination': new_path},
            user=user,
            timeout=timeout,
        )
        return normalize_entry((data or {}).get('entry'), extended=True)

    def watch_dir(self, path, recursive=False, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/filesystem.Filesystem/CreateWatcher',
            {'path': path, 'recursive': recursive},
            user=user,
            timeout=timeout,
        )
        watcher_id = (
            (data or {}).get('watcherId') or
            (data or {}).get('watcher_id')
        )
        return WatchHandle(self, watcher_id)

    watchDir = watch_dir



from .base import Middleware, compose_middleware
from .ua import UserAgentMiddleware
from .retry_domains import RetryDomainsMiddleware

__all__ = [
    'Middleware', 'compose_middleware',
    'UserAgentMiddleware',
    'RetryDomainsMiddleware'
]



import functools

from requests.adapters import HTTPAdapter

from qiniu import config, __version__

from .client import HTTPClient
from .middleware import UserAgentMiddleware

qn_http_client = HTTPClient(
    middlewares=[
        UserAgentMiddleware(__version__)
    ]
)


# compatibility with some config from qiniu.config
def _before_send(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        _init_http_adapter()
        return func(self, *args, **kwargs)

    return wrapper


qn_http_client.send_request = _before_send(qn_http_client.send_request)


def _init_http_adapter():
    # may be optimized:
    # only called when config changed, not every time before send request
    adapter = HTTPAdapter(
        pool_connections=config.get_default('connection_pool'),
        pool_maxsize=config.get_default('connection_pool'),
        max_retries=config.get_default('connection_retries'))
    qn_http_client.session.mount('http://', adapter)



# -*- coding: utf-8 -*-

from .client import SandboxClient
from .config import (
    env,
    load_dotenv_if_present,
    required_env,
    sandbox_client,
    sandbox_template,
)
from .commands import (
    CommandExitError,
    CommandHandle,
    CommandResult,
    Commands,
    ProcessInfo,
)
from .constants import (
    DEFAULT_ENDPOINT,
    DEFAULT_TEMPLATE,
    DEFAULT_USER,
    ENVD_PORT,
)
from .errors import (
    FileNotFoundException,
    GitAuthException,
    GitUpstreamException,
    InvalidArgumentException,
    SandboxError,
    TemplateBuildError,
)
from .filesystem import (
    EntryInfo,
    FileType,
    Filesystem,
    FilesystemEvent,
    FilesystemEventType,
    WatchHandle,
    WriteEntry,
    WriteInfo,
)
from .git import Git, GitBranches, GitFileStatus, GitStatus
from .pty import Pty, PtySize
from .resources import GitRepositoryResource, KodoResource
from .sandbox import Sandbox, SandboxPaginator
from .template import (
    ReadyCmd,
    Template,
    wait_for_file,
    wait_for_port,
    wait_for_process,
    wait_for_timeout,
    wait_for_url,
)

__all__ = [
    'CommandExitError',
    'CommandHandle',
    'CommandResult',
    'Commands',
    'DEFAULT_ENDPOINT',
    'DEFAULT_TEMPLATE',
    'DEFAULT_USER',
    'ENVD_PORT',
    'EntryInfo',
    'FileType',
    'FileNotFoundException',
    'Filesystem',
    'FilesystemEvent',
    'FilesystemEventType',
    'Git',
    'GitAuthException',
    'GitBranches',
    'GitFileStatus',
    'GitRepositoryResource',
    'GitStatus',
    'GitUpstreamException',
    'InvalidArgumentException',
    'KodoResource',
    'ProcessInfo',
    'Pty',
    'PtySize',
    'ReadyCmd',
    'Sandbox',
    'SandboxClient',
    'SandboxError',
    'SandboxPaginator',
    'Template',
    'TemplateBuildError',
    'WatchHandle',
    'WriteEntry',
    'WriteInfo',
    'env',
    'load_dotenv_if_present',
    'required_env',
    'sandbox_client',
    'sandbox_template',
    'wait_for_file',
    'wait_for_port',
    'wait_for_process',
    'wait_for_timeout',
    'wait_for_url',
]



# -*- coding: utf-8 -*-
import re
import uuid

from qiniu.compat import basestring

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from .errors import (
    CommandExitError,
    GitAuthException,
    GitUpstreamException,
    InvalidArgumentException,
    SandboxError,
)
from .util import shell_quote


def _normalize_paths(paths):
    if paths is None:
        return None
    if isinstance(paths, basestring):
        return [paths]
    return paths


RESET_MODES = set(['soft', 'mixed', 'hard', 'merge', 'keep'])


def _remove_credential_file(filesystem, path):
    if not path:
        return
    try:
        filesystem.remove(path)
    except Exception:
        pass


class _CredentialCleanupHandle(object):
    def __init__(self, handle, filesystem, path):
        self._handle = handle
        self._filesystem = filesystem
        self._path = path
        self._cleaned = [False]
        self.wait = self._wrap_wait(
            handle.wait, filesystem, path, self._cleaned)

    def __getattr__(self, name):
        return getattr(self._handle, name)

    @staticmethod
    def _cleanup(filesystem, path, cleaned):
        if not cleaned[0]:
            cleaned[0] = True
            _remove_credential_file(filesystem, path)

    @classmethod
    def _wrap_wait(cls, wait, filesystem, path, cleaned):
        def wrapped_wait(*args, **kwargs):
            try:
                return wait(*args, **kwargs)
            finally:
                cls._cleanup(filesystem, path, cleaned)
        return wrapped_wait

    def __del__(self):
        try:
            self._cleanup(self._filesystem, self._path, self._cleaned)
        except Exception:
            pass


def _looks_git_config_key(value):
    return bool(re.match(
        r'^[A-Za-z][A-Za-z0-9-]*(\.[A-Za-z][A-Za-z0-9-]*)+$',
        str(value or ''),
    ))


class GitFileStatus(object):
    def __init__(self, name, status, index_status, working_tree_status,
                 staged, renamed_from=None):
        self.name = name
        self.status = status
        self.index_status = index_status
        self.working_tree_status = working_tree_status
        self.staged = staged
        self.renamed_from = renamed_from


class GitStatus(object):
    def __init__(self, current_branch=None, upstream=None, ahead=0, behind=0,
                 detached=False, file_status=None):
        self.current_branch = current_branch
        self.upstream = upstream
        self.ahead = ahead
        self.behind = behind
        self.detached = detached
        self.file_status = file_status or []

    @property
    def is_clean(self):
        return len(self.file_status) == 0

    @property
    def has_changes(self):
        return len(self.file_status) > 0

    @property
    def has_staged(self):
        return any(item.staged for item in self.file_status)

    @property
    def has_untracked(self):
        return any(item.status == 'untracked' for item in self.file_status)

    @property
    def has_conflicts(self):
        return any(item.status == 'conflict' for item in self.file_status)

    @property
    def total_count(self):
        return len(self.file_status)

    @property
    def staged_count(self):
        return sum(1 for item in self.file_status if item.staged)

    @property
    def unstaged_count(self):
        return sum(1 for item in self.file_status if not item.staged)

    @property
    def untracked_count(self):
        return sum(1 for item in self.file_status
                   if item.status == 'untracked')

    @property
    def conflict_count(self):
        return sum(1 for item in self.file_status
                   if item.status == 'conflict')


class GitBranches(object):
    def __init__(self, branches=None, current_branch=None):
        self.branches = branches or []
        self.current_branch = current_branch


def _parse_ahead_behind(segment):
    ahead = 0
    behind = 0
    if not segment:
        return ahead, behind
    if 'ahead' in segment:
        try:
            ahead = int(segment.split('ahead')[1].split(',')[0].strip())
        except Exception:
            ahead = 0
    if 'behind' in segment:
        try:
            behind = int(segment.split('behind')[1].split(',')[0].strip())
        except Exception:
            behind = 0
    return ahead, behind


def _normalize_branch_name(name):
    if name.startswith('HEAD (detached at '):
        return name.replace('HEAD (detached at ', '').rstrip(')')
    return (
        name.replace('HEAD (no branch)', 'HEAD')
        .replace('No commits yet on ', '')
        .replace('Initial commit on ', '')
    )


def _derive_status(index_status, working_status):
    statuses = set([index_status, working_status])
    if 'U' in statuses:
        return 'conflict'
    if 'R' in statuses:
        return 'renamed'
    if 'C' in statuses:
        return 'copied'
    if 'D' in statuses:
        return 'deleted'
    if 'A' in statuses:
        return 'added'
    if 'M' in statuses:
        return 'modified'
    if 'T' in statuses:
        return 'typechange'
    if '?' in statuses:
        return 'untracked'
    return 'unknown'


def parse_git_status(output):
    lines = [line.rstrip() for line in (output or '').split('\n')
             if line.strip()]
    current_branch = None
    upstream = None
    ahead = 0
    behind = 0
    detached = False
    file_status = []

    if not lines:
        return GitStatus(file_status=file_status)

    branch_line = lines[0]
    if branch_line.startswith('## '):
        branch_info = branch_line[3:]
        ahead_start = branch_info.find(' [')
        branch_part = branch_info if ahead_start == -1 else branch_info[
            :ahead_start]
        ahead_part = None if ahead_start == -1 else branch_info[
            ahead_start + 2:-1]
        normalized = _normalize_branch_name(branch_part)
        is_detached = branch_part.startswith('HEAD (detached at ') or (
            'detached' in branch_part)
        if is_detached or normalized.startswith('HEAD'):
            detached = True
        elif '...' in normalized:
            current_branch, upstream = normalized.split('...', 1)
            current_branch = current_branch or None
            upstream = upstream or None
        else:
            current_branch = normalized or None
        ahead, behind = _parse_ahead_behind(ahead_part)

    for line in lines[1:]:
        if line.startswith('?? '):
            file_status.append(GitFileStatus(
                name=line[3:],
                status='untracked',
                index_status='?',
                working_tree_status='?',
                staged=False,
            ))
            continue
        if len(line) < 3:
            continue
        index_status = line[0]
        working_status = line[1]
        path = line[3:]
        renamed_from = None
        name = path
        if (index_status == 'R' or working_status == 'R') and ' -> ' in path:
            renamed_from, name = path.split(' -> ', 1)
        file_status.append(GitFileStatus(
            name=name,
            status=_derive_status(index_status, working_status),
            index_status=index_status,
            working_tree_status=working_status,
            staged=index_status not in (' ', '?'),
            renamed_from=renamed_from,
        ))

    return GitStatus(
        current_branch=current_branch,
        upstream=upstream,
        ahead=ahead,
        behind=behind,
        detached=detached,
        file_status=file_status,
    )


def parse_git_branches(output):
    branches = []
    current_branch = None
    for line in [line.strip() for line in (output or '').split('\n')
                 if line.strip()]:
        if '\t' in line:
            name, marker = line.split('\t', 1)
            marker = marker.strip()
        else:
            marker = '*' if line.startswith('* ') else ''
            name = line[2:] if line.startswith('* ') else line
        branches.append(name)
        if marker == '*':
            current_branch = name
    return GitBranches(branches=branches, current_branch=current_branch)


def _is_auth_failure(result):
    message = '{0}\n{1}\n{2}'.format(
        getattr(result, 'stderr', ''),
        getattr(result, 'stdout', ''),
        getattr(result, 'error', ''),
    ).lower()
    snippets = (
        'authentication failed',
        'terminal prompts disabled',
        'could not read username',
        'invalid username or password',
        'access denied',
        'permission denied',
        'not authorized',
    )
    return any(snippet in message for snippet in snippets)


def _is_missing_upstream(result):
    message = '{0}\n{1}\n{2}'.format(
        getattr(result, 'stderr', ''),
        getattr(result, 'stdout', ''),
        getattr(result, 'error', ''),
    ).lower()
    snippets = (
        'has no upstream branch',
        'no upstream branch',
        'no upstream configured',
        'no tracking information for the current branch',
        'no tracking information',
        'set the remote as upstream',
        'set the upstream branch',
        'please specify which branch you want to merge with',
    )
    return any(snippet in message for snippet in snippets)


def _validate_git_url_credentials(url, username, password):
    if not username and not password:
        return None
    if not username or not password:
        raise InvalidArgumentException(
            'Both username and password are required when using Git '
            'credentials.')
    parsed = urlparse(url)
    if parsed.scheme not in ('http', 'https'):
        raise InvalidArgumentException(
            'Only http(s) Git URLs support username/password credentials.')
    return parsed


def _askpass_script():
    return (
        '#!/bin/sh\n'
        'case "$1" in\n'
        '*Username*) printf "%s\\n" "$GIT_USERNAME" ;;\n'
        '*Password*) printf "%s\\n" "$GIT_PASSWORD" ;;\n'
        '*) printf "\\n" ;;\n'
        'esac\n'
    )


class Git(object):
    def __init__(self, commands):
        self.commands = commands

    def _run_git(self, repo_path, args, **opts):
        if repo_path:
            opts['cwd'] = repo_path
        return self.commands.run('git {0}'.format(' '.join(args)), **opts)

    def _get_remote_url(self, repo_path, remote, **opts):
        opts = dict(opts)
        opts.pop('background', None)
        result = self._run_git(
            repo_path,
            ['remote', 'get-url', shell_quote(remote)],
            **opts
        )
        url = (getattr(result, 'stdout', '') or '').strip()
        if not url:
            raise InvalidArgumentException(
                'Remote "{0}" URL not found in repository.'.format(remote))
        return url

    def _resolve_remote_name(self, repo_path, remote=None, **opts):
        if remote:
            return remote
        opts = dict(opts)
        opts.pop('background', None)
        result = self._run_git(repo_path, ['remote'], **opts)
        remotes = [
            line.strip()
            for line in (getattr(result, 'stdout', '') or '').splitlines()
            if line.strip()
        ]
        if len(remotes) == 1:
            return remotes[0]
        if len(remotes) == 0:
            raise InvalidArgumentException(
                'No remotes found in the repository.')
        raise InvalidArgumentException(
            'Remote is required when using username/password and the '
            'repository has multiple remotes.')

    def _with_remote_credentials(self, repo_path, remote, username, password,
                                 operation, **opts):
        original_url = self._get_remote_url(repo_path, remote, **opts)
        _validate_git_url_credentials(original_url, username, password)
        sandbox = getattr(self.commands, 'sandbox', None)
        filesystem = getattr(sandbox, 'files', None)
        if filesystem is None:
            raise SandboxError(
                'Sandbox filesystem is required for credentialed Git '
                'operations.')
        temp_dir = '/tmp/qiniu-git-auth'
        setup_opts = dict(opts)
        setup_opts.pop('background', None)
        prepare_result = self.commands.run(
            'install -d -m 700 {0}'.format(shell_quote(temp_dir)),
            **setup_opts
        )
        if getattr(prepare_result, 'exit_code', 0):
            return prepare_result
        askpass_path = '{0}/qiniu-git-askpass-{1}'.format(
            temp_dir,
            uuid.uuid4().hex,
        )
        try:
            filesystem.write(askpass_path, _askpass_script())
            chmod_result = self.commands.run(
                'chmod 700 {0}'.format(shell_quote(askpass_path)),
                **setup_opts
            )
            if getattr(chmod_result, 'exit_code', 0):
                _remove_credential_file(filesystem, askpass_path)
                return chmod_result

            auth_opts = dict(opts)
            envs = dict(auth_opts.get('envs') or {})
            envs.update({
                'GIT_ASKPASS': askpass_path,
                'GIT_TERMINAL_PROMPT': '0',
                'GIT_USERNAME': username,
                'GIT_PASSWORD': password,
            })
            auth_opts['envs'] = envs

            result = operation(auth_opts)
            if opts.get('background') and hasattr(result, 'wait'):
                result = _CredentialCleanupHandle(
                    result, filesystem, askpass_path)
            else:
                _remove_credential_file(filesystem, askpass_path)
            return result
        except BaseException as err:
            _remove_credential_file(filesystem, askpass_path)
            raise err

    def _raise_known_result_error(
            self, result, operation, throw_on_error=False):
        if (getattr(result, 'exit_code', None) == -1 and
                hasattr(result, 'wait')):
            if throw_on_error:
                result.wait = self._raise_after_wait(result.wait, operation)
            return
        if not hasattr(result, 'exit_code'):
            return
        if result.exit_code:
            if _is_auth_failure(result):
                raise GitAuthException(
                    'Git {0} requires credentials for private repositories.'
                    .format(operation))
            if _is_missing_upstream(result):
                raise GitUpstreamException(
                    'Git {0} failed because no upstream branch is configured.'
                    .format(operation))
            if throw_on_error:
                raise CommandExitError(result)

    def _raise_after_wait(self, wait, operation):
        def wrapped_wait(*args, **kwargs):
            result = wait(*args, **kwargs)
            self._raise_known_result_error(
                result, operation, throw_on_error=True)
            return result
        return wrapped_wait

    def clone(self, repo_url, path=None, branch=None, depth=None, **opts):
        args = ['clone']
        if depth:
            args.extend(['--depth', shell_quote(depth)])
        if branch:
            args.extend(['--branch', shell_quote(branch)])
        args.append(shell_quote(repo_url))
        if path:
            args.append(shell_quote(path))
        return self._run_git(None, args, **opts)

    def init(self, repo_path, bare=False, initial_branch=None, **opts):
        args = ['init']
        if bare:
            args.append('--bare')
        if initial_branch:
            args.extend(['--initial-branch', shell_quote(initial_branch)])
        return self._run_git(repo_path, args, **opts)

    def status(self, repo_path, **opts):
        result = self._run_git(
            repo_path, [
                'status', '--porcelain=1', '-b'], **opts)
        if result.exit_code:
            raise CommandExitError(result)
        return parse_git_status(result.stdout)

    def add(self, repo_path, files=None, all=False, **opts):
        args = ['add']
        if all:
            args.append('--all')
        else:
            args.append('--')
            for path in _normalize_paths(files) or ['.']:
                args.append(shell_quote(path))
        return self._run_git(repo_path, args, **opts)

    def commit(self, repo_path, message, author_name=None, author_email=None,
               allow_empty=False, **opts):
        args = []
        if author_name:
            args.extend(['-c', shell_quote('user.name={0}'.format(
                author_name))])
        if author_email:
            args.extend(['-c', shell_quote('user.email={0}'.format(
                author_email))])
        args.extend(['commit', '-m', shell_quote(message)])
        if allow_empty:
            args.append('--allow-empty')
        return self._run_git(repo_path, args, **opts)

    def configure_user(self, repo_path, name, email, **opts):
        name_result = self._run_git(
            repo_path,
            ['config', 'user.name', shell_quote(name)],
            **opts
        )
        if name_result.exit_code != 0:
            return name_result
        return self._run_git(
            repo_path,
            ['config', 'user.email', shell_quote(email)],
            **opts
        )

    configureUser = configure_user

    def pull(self, repo_path, remote=None, branch=None, username=None,
             password=None, **opts):
        throw_on_error = opts.pop('throw_on_error', False)
        if password and not username:
            raise InvalidArgumentException(
                'Git pull requires username when password is provided')
        remote_name = None
        if username and password:
            remote_name = self._resolve_remote_name(repo_path, remote, **opts)

        args = ['pull']
        target_remote = remote_name or remote
        if target_remote:
            args.append(shell_quote(target_remote))
        if branch:
            args.append(shell_quote(branch))
        if username and password:
            result = self._with_remote_credentials(
                repo_path,
                remote_name,
                username,
                password,
                lambda auth_opts: self._run_git(
                    repo_path, args, throw_on_error=False, **auth_opts),
                **opts
            )
            self._raise_known_result_error(
                result, 'pull', throw_on_error=throw_on_error)
            return result

        result = self._run_git(repo_path, args, throw_on_error=False, **opts)
        self._raise_known_result_error(
            result, 'pull', throw_on_error=throw_on_error)
        return result

    def push(self, repo_path, remote=None, branch=None, set_upstream=True,
             username=None, password=None, **opts):
        throw_on_error = opts.pop('throw_on_error', False)
        if password and not username:
            raise InvalidArgumentException(
                'Git push requires username when password is provided')
        remote_name = None
        if username and password:
            remote_name = self._resolve_remote_name(repo_path, remote, **opts)

        args = ['push']
        target_remote = remote_name or remote
        if set_upstream and target_remote:
            args.append('--set-upstream')
        if target_remote:
            args.append(shell_quote(target_remote))
        if branch:
            args.append(shell_quote(branch))
        if username and password:
            result = self._with_remote_credentials(
                repo_path,
                remote_name,
                username,
                password,
                lambda auth_opts: self._run_git(
                    repo_path, args, throw_on_error=False, **auth_opts),
                **opts
            )
            self._raise_known_result_error(
                result, 'push', throw_on_error=throw_on_error)
            return result

        result = self._run_git(repo_path, args, throw_on_error=False, **opts)
        self._raise_known_result_error(
            result, 'push', throw_on_error=throw_on_error)
        return result

    def dangerously_authenticate(
            self,
            username,
            password,
            host='github.com',
            protocol='https',
            **opts):
        opts = dict(opts)
        opts.pop('background', None)
        if not username:
            raise InvalidArgumentException('username is required')
        if not password:
            raise InvalidArgumentException('password is required')

        result = self._run_git(
            None,
            ['config', '--global', 'credential.helper', 'store'],
            **opts
        )
        if result.exit_code != 0:
            return result

        credential = (
            'protocol={0}\n'
            'host={1}\n'
            'username={2}\n'
            'password={3}\n\n'
        ).format(protocol, host, username, password)
        handle = self.commands.run(
            'git credential approve',
            stdin=True,
            background=True,
            **opts
        )
        self.commands.send_stdin(handle.pid, credential)
        self.commands.close_stdin(handle.pid)
        return handle.wait()

    dangerouslyAuthenticate = dangerously_authenticate

    def remote_add(self, repo_path, name, url, **opts):
        fetch = opts.pop('fetch', False)
        overwrite = opts.pop('overwrite', False)
        if overwrite:
            self._run_git(repo_path, [
                'remote',
                'remove',
                shell_quote(name),
            ], **opts)
        result = self._run_git(repo_path, [
            'remote',
            'add',
            shell_quote(name),
            shell_quote(url),
        ], **opts)
        if result.exit_code != 0 or not fetch:
            return result
        return self._run_git(repo_path, [
            'fetch',
            shell_quote(name),
        ], **opts)

    remoteAdd = remote_add

    def remote_get(self, repo_path, name='origin', **opts):
        return self._run_git(repo_path, [
            'remote',
            'get-url',
            shell_quote(name),
        ], **opts)

    remoteGet = remote_get

    def branches(self, repo_path, **opts):
        result = self._run_git(
            repo_path,
            ['branch', shell_quote('--format=%(refname:short)\t%(HEAD)')],
            **opts
        )
        return parse_git_branches(result.stdout)

    def create_branch(self, repo_path, name, start_point=None, **opts):
        args = ['checkout', '-b', shell_quote(name)]
        if start_point:
            args.append(shell_quote(start_point))
        return self._run_git(repo_path, args, **opts)

    createBranch = create_branch

    def checkout_branch(self, repo_path, name, create=False, **opts):
        args = ['checkout']
        if create:
            args.append('-b')
        args.append(shell_quote(name))
        return self._run_git(repo_path, args, **opts)

    checkoutBranch = checkout_branch

    def delete_branch(self, repo_path, name, force=True, **opts):
        return self._run_git(repo_path, [
            'branch',
            '-D' if force else '-d',
            shell_quote(name),
        ], **opts)

    deleteBranch = delete_branch

    def reset(self, repo_path, target=None, mode=None, **opts):
        args = ['reset']
        mode = mode or opts.get('reset_type') or opts.get('resetType')
        if mode:
            if mode not in RESET_MODES:
                raise InvalidArgumentException(
                    'Unsupported git reset mode: {0}'.format(mode)
                )
            args.append('--{0}'.format(mode))
        if target:
            args.append(shell_quote(target))
        return self._run_git(repo_path, args, **opts)

    def restore(self, repo_path, paths=None, staged=False, source=None,
                **opts):
        args = ['restore']
        if staged:
            args.append('--staged')
        if source:
            args.extend(['--source', shell_quote(source)])
        paths = paths if paths is not None else opts.get('files')
        args.append('--')
        for path in _normalize_paths(paths) or ['.']:
            args.append(shell_quote(path))
        return self._run_git(repo_path, args, **opts)

    def set_config(self, key, value, scope='global', path=None, **opts):
        """Set a Git config value.

        Preferred signature:
            set_config(key, value, scope='global', path=None, **opts)

        For local config, pass scope='local' and path=repo_path. For backward
        compatibility this method also accepts the deprecated signature:
            set_config(repo_path, key, value, global_config=False, **opts)
        """
        key, value, scope, path = self._normalize_set_config_args(
            key, value, scope, path, opts)
        scope_flag, repo_path = self._resolve_config_scope(scope, path)
        args = ['config']
        if scope_flag:
            args.append(scope_flag)
        args.extend([shell_quote(key), shell_quote(value)])
        return self._run_git(repo_path, args, **opts)

    setConfig = set_config

    def get_config(self, key, scope='global', path=None, **opts):
        """Get a Git config value.

        Preferred signature:
            get_config(key, scope='global', path=None, **opts)

        For local config, pass scope='local' and path=repo_path. For backward
        compatibility this method also accepts the deprecated signature:
            get_config(repo_path, key, global_config=False, **opts)
        """
        key, scope, path = self._normalize_get_config_args(
            key, scope, path, opts)
        scope_flag, repo_path = self._resolve_config_scope(scope, path)
        args = ['config']
        if scope_flag:
            args.append(scope_flag)
        args.extend(['--get', shell_quote(key)])
        return self._run_git(repo_path, args, **opts)

    getConfig = get_config

    def _normalize_set_config_args(self, key, value, scope, path, opts):
        global_config = opts.pop(
            'global_config', opts.pop('globalConfig', False))
        if global_config or self._is_legacy_config_call(key, value, scope):
            repo_path = key
            key = value
            value = scope
            scope = 'global' if global_config or repo_path is None else 'local'
            path = None if scope == 'global' else repo_path
        return key, value, scope, path

    def _normalize_get_config_args(self, key, scope, path, opts):
        global_config = opts.pop(
            'global_config', opts.pop('globalConfig', False))
        if global_config or self._is_legacy_config_call(key, None, scope):
            repo_path = key
            key = scope
            scope = 'global' if global_config or repo_path is None else 'local'
            path = None if scope == 'global' else repo_path
        return key, scope, path

    def _is_legacy_config_call(self, key, value, scope):
        if scope is None:
            return False
        scope_name = str(scope).strip().lower()
        if scope_name not in ('global', 'local', 'system'):
            return True
        key_text = str(key or '')
        return (
            '/' in key_text or
            '\\' in key_text or
            key_text in ('.', '..') or
            not _looks_git_config_key(key_text) or
            _looks_git_config_key(value)
        )

    def _resolve_config_scope(self, scope=None, path=None):
        scope_name = (scope or 'global').strip().lower()
        if scope_name not in ('global', 'local', 'system'):
            raise InvalidArgumentException(
                'Git config scope must be global, local, or system')
        if scope_name == 'local':
            if not path:
                raise InvalidArgumentException(
                    'Repository path is required for local scope')
            return '--local', path
        if scope_name == 'system':
            return '--system', None
        return '--global', None



import threading


class _FlightLock:
    """
    Do not use dataclass which caused the event created only once
    """
    def __init__(self):
        self.event = threading.Event()
        self.result = None
        self.error = None


class SingleFlight:
    def __init__(self):
        self._locks = {}
        self._lock = threading.Lock()

    def do(self, key, fn, *args, **kwargs):
        # here does not use `with` statement
        # because need to wait by another object if it exists,
        # and reduce the `acquire` times if it not exists
        self._lock.acquire()
        if key in self._locks:
            flight_lock = self._locks[key]

            self._lock.release()
            flight_lock.event.wait()

            if flight_lock.error:
                raise flight_lock.error
            return flight_lock.result

        flight_lock = _FlightLock()
        self._locks[key] = flight_lock
        self._lock.release()

        try:
            flight_lock.result = fn(*args, **kwargs)
        except Exception as e:
            flight_lock.error = e
        finally:
            flight_lock.event.set()

        with self._lock:
            del self._locks[key]

        if flight_lock.error:
            raise flight_lock.error
        return flight_lock.result



from qiniu.retry.abc import RetryPolicy

from .region import Region, ServiceName


class RegionsRetryPolicy(RetryPolicy):
    def __init__(
        self,
        regions_provider,
        service_names,
        preferred_endpoints_provider=None,
        on_change_region=None
    ):
        """
        Parameters
        ----------
        regions_provider: Iterable[Region]
        service_names: list[ServiceName or str]
        preferred_endpoints_provider: Iterable[Endpoint]
        on_change_region: Callable
            `(context: dict) -> None`
        """
        self.regions_provider = regions_provider
        self.service_names = service_names
        if not service_names:
            raise ValueError('Must provide at least one service name')
        if preferred_endpoints_provider is None:
            preferred_endpoints_provider = []
        self.preferred_endpoints_provider = preferred_endpoints_provider
        self.on_change_region = on_change_region

    def init_context(self, context):
        """
        Parameters
        ----------
        context: dict
        """
        self._init_regions(context)
        self._prepare_endpoints(context)

    def should_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: Attempt
        """
        return (
            len(attempt.context.get('alternative_regions', [])) > 0 or
            len(attempt.context.get('alternative_service_names', [])) > 0
        )

    def prepare_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: Attempt
        """
        if attempt.context.get('alternative_service_names'):
            # change service for next try
            attempt.context['service_name'] = attempt.context.get('alternative_service_names').pop(0)
        elif attempt.context.get('alternative_regions'):
            # change region for next try
            attempt.context['region'] = attempt.context.get('alternative_regions').pop(0)
            if callable(self.on_change_region):
                self.on_change_region(attempt.context)
        else:
            raise RuntimeError('There isn\'t available region or service for next try')
        self._prepare_endpoints(attempt.context)

    def _init_regions(self, context):
        """
        Parameters
        ----------
        context: dict
        """
        regions = list(self.regions_provider)
        preferred_endpoints = list(self.preferred_endpoints_provider)
        if not regions and not preferred_endpoints:
            raise ValueError('There isn\'t available region or preferred endpoint')

        if not preferred_endpoints:
            # regions are not empty implicitly by above if condition
            context['alternative_regions'] = regions
            context['region'] = context['alternative_regions'].pop(0)
            # shallow copy list
            # change to `list.copy` for more readable when min version of python update to >= 3
            context['alternative_service_names'] = self.service_names[:]
            context['service_name'] = context['alternative_service_names'].pop(0)
            return

        # find preferred service name and region by preferred endpoints
        preferred_region_index = -1
        preferred_service_index = -1
        for ri, region in enumerate(regions):
            for si, service_name in enumerate(self.service_names):
                if any(
                    pe.host in [
                        e.host for e in region.services.get(service_name, [])
                    ]
                    for pe in preferred_endpoints
                ):
                    preferred_region_index = ri
                    preferred_service_index = si
                    break

        # initialize the order of service_names and regions
        if preferred_region_index < 0:
            # shallow copy list
            # change to `list.copy` for more readable when min version of python update to >= 3
            context['alternative_service_names'] = self.service_names[:]
            context['service_name'] = context['alternative_service_names'].pop(0)

            context['region'] = Region(
                region_id='preferred_region',
                services={
                    context['service_name']: preferred_endpoints
                }
            )
            context['alternative_regions'] = regions
        else:
            # regions are not empty implicitly by above if condition
            # preferred endpoints are in a known region, then reorder the regions and services
            context['alternative_regions'] = regions
            context['region'] = context['alternative_regions'].pop(preferred_region_index)
            # shallow copy list
            # change to `list.copy` for more readable when min version of python update to >= 3
            context['alternative_service_names'] = self.service_names[:]
            context['service_name'] = context['alternative_service_names'].pop(preferred_service_index)

    def _prepare_endpoints(self, context):
        """
        Parameters
        ----------
        context: dict
        """
        # shallow copy list
        # change to `list.copy` for more readable when min version of python update to >= 3
        endpoints = context['region'].services.get(context['service_name'], [])[:]
        while not endpoints:
            if context['alternative_service_names']:
                context['service_name'] = context['alternative_service_names'].pop(0)
                endpoints = context['region'].services.get(context['service_name'], [])[:]
            elif context['alternative_regions']:
                context['region'] = context['alternative_regions'].pop(0)
                # shallow copy list
                # change to `list.copy` for more readable when min version of python update to >= 3
                context['alternative_service_names'] = self.service_names[:]
                context['service_name'] = context['alternative_service_names'].pop(0)
                endpoints = context['region'].services.get(context['service_name'], [])[:]
                if callable(self.on_change_region):
                    self.on_change_region(context)
            else:
                raise RuntimeError(
                    'There isn\'t available endpoint for {0} service(s) in any available regions'.format(
                        ', '.join(
                            sn.value if isinstance(sn, ServiceName) else sn
                            for sn in self.service_names
                        )
                    )
                )
        context['alternative_endpoints'] = endpoints
        context['endpoint'] = context['alternative_endpoints'].pop(0)



# -*- coding: utf-8 -*-
import hashlib
import json
import os
import tempfile
from qiniu.compat import is_py2


class UploadProgressRecorder(object):
    """
    持久化上传记录类

    该类默认保存每个文件的上传记录到文件系统中，用于断点续传
    上传记录为json格式

    Attributes:
        record_folder: 保存上传记录的目录
    """

    def __init__(self, record_folder=tempfile.gettempdir()):
        self.record_folder = record_folder

    def __get_upload_record_file_path(self, file_name, key):
        record_key = '{0}/{1}'.format(key, file_name)
        if is_py2:
            record_file_name = hashlib.md5(record_key).hexdigest()
        else:
            record_file_name = hashlib.md5(record_key.encode('utf-8')).hexdigest()
        return os.path.join(self.record_folder, record_file_name)

    def has_upload_record(self, file_name, key):
        upload_record_file_path = self.__get_upload_record_file_path(file_name, key)
        return os.path.isfile(upload_record_file_path)

    def get_upload_record(self, file_name, key):
        upload_record_file_path = self.__get_upload_record_file_path(file_name, key)
        if not self.has_upload_record(file_name, key):
            return None
        try:
            with open(upload_record_file_path, 'r') as f:
                json_data = json.load(f)
        except (IOError, ValueError):
            json_data = None

        return json_data

    def set_upload_record(self, file_name, key, data):
        upload_record_file_path = self.__get_upload_record_file_path(file_name, key)
        with open(upload_record_file_path, 'w') as f:
            json.dump(data, f)

    def delete_upload_record(self, file_name, key):
        upload_record_file_path = self.__get_upload_record_file_path(file_name, key)
        try:
            os.remove(upload_record_file_path)
        except OSError:
            pass



# -*- coding: utf-8 -*-
from qiniu import http, Auth
import json


class RtcServer(object):
    """
    直播连麦管理类
    主要涉及了直播连麦管理及操作接口的实现，具体的接口规格可以参考官方文档 https://developer.qiniu.com
    Attributes:
        auth: 账号管理密钥对，Auth对象

    """

    def __init__(self, auth):
        self.auth = auth
        self.host = 'http://rtc.qiniuapi.com'

    def create_app(self, data):
        return self.__post(self.host + '/v3/apps', data)

    def get_app(self, app_id=None):
        if app_id:
            return self.__get(self.host + '/v3/apps/%s' % app_id)
        else:
            return self.__get(self.host + '/v3/apps')

    def delete_app(self, app_id):
        return self.__delete(self.host + '/v3/apps/%s' % app_id)

    def update_app(self, app_id, data):
        return self.__post(self.host + '/v3/apps/%s' % app_id, data)

    def list_user(self, app_id, room_name):
        return self.__get(self.host + '/v3/apps/%s/rooms/%s/users' % (app_id, room_name))

    def kick_user(self, app_id, room_name, user_id):
        return self.__delete(self.host + '/v3/apps/%s/rooms/%s/users/%s' % (app_id, room_name, user_id))

    def list_active_rooms(self, app_id, room_name_prefix=None):
        if room_name_prefix:
            return self.__get(self.host + '/v3/apps/%s/rooms?prefix=%s' % (app_id, room_name_prefix))
        else:
            return self.__get(self.host + '/v3/apps/%s/rooms' % app_id)

    def __post(self, url, data=None):
        return http._post_with_qiniu_mac(url, data, self.auth)

    def __get(self, url, params=None):
        return http._get_with_qiniu_mac(url, params, self.auth)

    def __delete(self, url, params=None):
        return http._delete_with_qiniu_mac(url, params, self.auth)


def get_room_token(access_key, secret_key, room_access):
    auth = Auth(access_key, secret_key)
    room_access_str = json.dumps(room_access)
    room_token = auth.token_with_data(room_access_str)
    return room_token



# -*- coding: utf-8 -*-
import logging

import requests

from qiniu.config import get_default
from .response import ResponseInfo
from .middleware import compose_middleware


class HTTPClient:
    def __init__(self, middlewares=None, send_opts=None):
        self.session = requests.Session()
        self.middlewares = [] if middlewares is None else middlewares
        self.send_opts = {} if send_opts is None else send_opts

    def _wrap_send(self, req, **kwargs):
        # compatibility with setting timeout by qiniu.config.set_default
        kwargs.setdefault('timeout', get_default('connection_timeout'))

        resp = self.session.send(req.prepare(), **kwargs)
        return ResponseInfo(resp, None)

    def send_request(self, request, middlewares=None, **kwargs):
        """

        Args:
            request (requests.Request):
                requests.Request 对象

            middlewares (list[qiniu.http.middleware.Middleware] or (list[qiniu.http.middleware.Middleware]) -> list[qiniu.http.middleware.Middleware]):
                仅对本次请求生效的中间件。

                如果传入的是列表，那么会作为追加的中间件拼接到 Client 中间件的后面。

                也可传入函数，获得 Client 中间件的一个副本来做更细的控制。
                例如拼接到 Client 中间件的前面，可以这样使用：

                c.send_request(my_req, middlewares=lambda mws: my_mws + mws)

            kwargs:
                将作为其他参数直接透传给 session.send 方法


        Returns:
            (dict, ResponseInfo): 可拆包的一个元组。
            第一个元素为响应体的 dict，若响应体为 json 的话。
            第二个元素为包装过的响应内容，包括了更多的响应内容。

        """

        # set default values
        middlewares = [] if middlewares is None else middlewares

        # join middlewares and client middlewares
        mw_ls = []
        if callable(middlewares):
            mw_ls = middlewares(self.middlewares.copy())
        elif isinstance(middlewares, list):
            mw_ls = self.middlewares + middlewares

        # send request
        try:
            handle = compose_middleware(
                mw_ls,
                lambda req: self._wrap_send(req, **kwargs)
            )
            resp_info = handle(request)
        except Exception as e:
            return None, ResponseInfo(None, e)

        # if ok try dump response info to dict from json
        if not resp_info.ok():
            return None, resp_info

        try:
            ret = resp_info.json()
        except ValueError:
            logging.debug("response body decode error: %s" % resp_info.text_body)
            ret = {}
        return ret, resp_info

    def get(
        self,
        url,
        params=None,
        auth=None,
        headers=None,
        middlewares=None,
        **kwargs
    ):
        req = requests.Request(
            method='get',
            url=url,
            params=params,
            auth=auth,
            headers=headers
        )
        send_opts = self.send_opts.copy()
        send_opts.update(kwargs)
        send_opts.setdefault("allow_redirects", True)
        return self.send_request(
            req,
            middlewares=middlewares,
            **send_opts
        )

    def post(
        self,
        url,
        data,
        files,
        auth=None,
        headers=None,
        middlewares=None,
        **kwargs
    ):
        req = requests.Request(
            method='post',
            url=url,
            data=data,
            files=files,
            auth=auth,
            headers=headers
        )
        send_opts = self.send_opts.copy()
        send_opts.update(kwargs)
        return self.send_request(
            req,
            middlewares=middlewares,
            **send_opts
        )

    def put(
        self,
        url,
        data,
        files,
        auth=None,
        headers=None,
        middlewares=None,
        **kwargs
    ):
        req = requests.Request(
            method='put',
            url=url,
            data=data,
            files=files,
            auth=auth,
            headers=headers
        )
        send_opts = self.send_opts.copy()
        send_opts.update(kwargs)
        return self.send_request(
            req,
            middlewares=middlewares,
            **send_opts
        )

    def delete(
        self,
        url,
        params,
        auth=None,
        headers=None,
        middlewares=None,
        **kwargs
    ):
        req = requests.Request(
            method='delete',
            url=url,
            params=params,
            auth=auth,
            headers=headers
        )
        send_opts = self.send_opts.copy()
        send_opts.update(kwargs)
        return self.send_request(
            req,
            middlewares=middlewares,
            **send_opts
        )



# -*- coding: utf-8 -*-
import base64
import binascii
import re

from qiniu.compat import basestring, bytes as bytes_type, str as text_type

from .envd import connect_rpc, connect_stream_rpc
from .errors import CommandExitError, SandboxError


class ProcessInfo(object):
    def __init__(self, pid=None, tag=None, cmd=None, args=None, envs=None,
                 cwd=None):
        self.pid = pid
        self.tag = tag
        self.cmd = cmd
        self.args = args or []
        self.envs = envs or {}
        self.cwd = cwd

    def to_dict(self):
        return {
            'pid': self.pid,
            'tag': self.tag,
            'cmd': self.cmd,
            'args': self.args,
            'envs': self.envs,
            'cwd': self.cwd,
        }


class CommandResult(object):
    def __init__(self, pid=0, exit_code=0, stdout='', stderr='', error=''):
        self.pid = pid
        self.exit_code = exit_code
        self.exitCode = exit_code
        self.stdout = stdout or ''
        self.stderr = stderr or ''
        self.error = error or ''

    def to_dict(self):
        return {
            'pid': self.pid,
            'exitCode': self.exit_code,
            'stdout': self.stdout,
            'stderr': self.stderr,
            'error': self.error,
        }


def _decode_bytes(value):
    if value is None:
        return ''
    if isinstance(value, list):
        return bytearray(value).decode('utf-8', 'replace')
    if isinstance(value, bytes_type):
        return value.decode('utf-8', 'replace')
    if isinstance(value, basestring):
        try:
            return base64.b64decode(value).decode('utf-8', 'replace')
        except (TypeError, ValueError, binascii.Error):
            return text_type(value)
    return text_type(value)


def command_result_from_events(events, on_stdout=None, on_stderr=None,
                               on_pty=None):
    pid = 0
    stdout = ''
    stderr = ''
    exit_code = -1
    error = ''
    for raw in events or []:
        event = raw.get('event', raw) if isinstance(raw, dict) else {}
        start = event.get('start')
        data = event.get('data')
        end = event.get('end')
        if start:
            pid = start.get('pid') or pid
        if data:
            stdout_chunk = _decode_bytes(data.get('stdout'))
            stderr_chunk = _decode_bytes(data.get('stderr'))
            pty_chunk = _decode_bytes(data.get('pty'))
            if stdout_chunk and on_stdout:
                on_stdout(stdout_chunk)
            if stderr_chunk and on_stderr:
                on_stderr(stderr_chunk)
            if pty_chunk and on_pty:
                on_pty(pty_chunk)
            stdout += stdout_chunk
            stderr += stderr_chunk
            stdout += pty_chunk
        if end:
            if end.get('exitCode') is not None:
                exit_code = end.get('exitCode')
            elif end.get('exit_code') is not None:
                exit_code = end.get('exit_code')
            else:
                status_match = re.match(
                    r'^exit status (\d+)$',
                    text_type(end.get('status') or ''),
                )
                if status_match:
                    exit_code = int(status_match.group(1))
            error = end.get('error') or ''
    return CommandResult(pid, exit_code, stdout, stderr, error)


class CommandHandle(object):
    def __init__(self, commands, result=None, throw_on_error=False,
                 events=None, on_stdout=None, on_stderr=None):
        self.commands = commands
        result = result or CommandResult()
        self.result = result
        self.pid = result.pid
        self.exit_code = result.exit_code
        self.exitCode = result.exit_code
        self.stdout = result.stdout
        self.stderr = result.stderr
        self.throw_on_error = throw_on_error
        self._events = events
        self._on_stdout = on_stdout
        self._on_stderr = on_stderr

    def wait(self, on_stdout=None, on_stderr=None):
        if self._events is not None:
            result = command_result_from_events(
                self._events,
                on_stdout=on_stdout or self._on_stdout,
                on_stderr=on_stderr or self._on_stderr,
            )
            if not result.pid:
                result.pid = self.pid
            result.stdout = self.result.stdout + result.stdout
            result.stderr = self.result.stderr + result.stderr
            if not result.error:
                result.error = self.result.error
            self.result = result
            self.pid = result.pid
            self.exit_code = result.exit_code
            self.exitCode = result.exit_code
            self.stdout = result.stdout
            self.stderr = result.stderr
            self._events = None
        if self.throw_on_error and self.result.exit_code:
            raise CommandExitError(self.result)
        return self.result

    def kill(self):
        return self.commands.kill(self.pid)


class Commands(object):
    def __init__(self, sandbox):
        self.sandbox = sandbox

    def run(self, cmd, cwd=None, envs=None, user=None, stdin=False,
            tag=None, background=False, throw_on_error=False,
            timeout=None, request_timeout=None, on_stdout=None,
            on_stderr=None, **opts):
        handle = self.start(
            cmd,
            cwd=cwd,
            envs=envs,
            user=user,
            stdin=stdin,
            tag=tag,
            throw_on_error=throw_on_error,
            timeout=timeout,
            request_timeout=request_timeout,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
            **opts
        )
        return handle if background else handle.wait()

    def start(self, cmd, cwd=None, envs=None, user=None, stdin=False,
              tag=None, throw_on_error=False, timeout=None,
              request_timeout=None, on_stdout=None, on_stderr=None, **opts):
        process = {
            'cmd': '/bin/bash',
            'args': ['-l', '-c', cmd],
        }
        if cwd:
            process['cwd'] = cwd
        if envs:
            process['envs'] = envs
        body = {
            'process': process,
            'stdin': stdin,
        }
        if tag:
            body['tag'] = tag
        if timeout is not None:
            body['timeout'] = timeout
        stream_timeout = request_timeout
        if stream_timeout is None:
            stream_timeout = timeout
        events = connect_stream_rpc(
            self.sandbox,
            '/process.Process/Start',
            body,
            user=user,
            timeout=stream_timeout,
            stream=True,
        )
        events = iter(events)
        try:
            first_event = next(events)
        except StopIteration:
            first_event = None
        result = command_result_from_events(
            [first_event] if first_event else [],
            on_stdout=on_stdout,
            on_stderr=on_stderr,
        )
        return CommandHandle(
            self,
            result,
            throw_on_error=throw_on_error,
            events=events,
            on_stdout=on_stdout,
            on_stderr=on_stderr,
        )

    def list(self, user=None, timeout=None):
        data = connect_rpc(
            self.sandbox,
            '/process.Process/List',
            {},
            user=user,
            timeout=timeout)
        processes = data.get('processes', []) if isinstance(data, dict) else []
        result = []
        for process in processes:
            config = process.get('config') or {}
            result.append(ProcessInfo(
                pid=process.get('pid'),
                tag=process.get('tag'),
                cmd=config.get('cmd'),
                args=config.get('args'),
                envs=config.get('envs'),
                cwd=config.get('cwd'),
            ))
        return result

    def connect(self, pid, tag=None, user=None, timeout=None,
                throw_on_error=False):
        selector = {'pid': pid}
        if tag is not None:
            selector = {'tag': tag}
        events = connect_stream_rpc(
            self.sandbox,
            '/process.Process/Connect',
            {'process': {'selector': selector}},
            user=user,
            timeout=timeout,
            stream=True,
        )
        events = iter(events)
        try:
            first_event = next(events)
        except StopIteration:
            first_event = None
        result = command_result_from_events([first_event])
        return CommandHandle(
            self,
            result,
            events=events,
            throw_on_error=throw_on_error,
        )

    def send_stdin(self, pid, data, user=None, timeout=None):
        if isinstance(data, bytearray):
            data = bytes(data)
        elif not isinstance(data, bytes):
            if hasattr(data, 'encode'):
                data = data.encode('utf-8')
            else:
                data = str(data).encode('utf-8')
        return connect_rpc(self.sandbox, '/process.Process/SendInput', {
            'process': {'selector': {'pid': pid}},
            'input': {'stdin': base64.b64encode(data).decode('ascii')},
        }, user=user, timeout=timeout)

    sendStdin = send_stdin

    def close_stdin(self, pid, user=None, timeout=None):
        return connect_rpc(self.sandbox, '/process.Process/CloseStdin', {
            'process': {'selector': {'pid': pid}},
        }, user=user, timeout=timeout)

    closeStdin = close_stdin

    def kill(self, pid, user=None, timeout=None):
        try:
            connect_rpc(self.sandbox, '/process.Process/SendSignal', {
                'process': {'selector': {'pid': pid}},
                'signal': 'SIGNAL_SIGKILL',
            }, user=user, timeout=timeout)
            return True
        except SandboxError as err:
            if getattr(err, 'status_code', None) == 404:
                return False
            raise



from qiniu.retry.abc import RetryPolicy


class EndpointsRetryPolicy(RetryPolicy):
    def __init__(self, endpoints_provider=None, skip_init_context=False):
        """
        Parameters
        ----------
        endpoints_provider: Iterable[Endpoint]
        skip_init_context: bool
        """
        self.endpoints_provider = endpoints_provider if endpoints_provider else []
        self.skip_init_context = skip_init_context

    def init_context(self, context):
        """
        Parameters
        ----------
        context: dict

        Returns
        -------
        None
        """
        if self.skip_init_context:
            return
        context['alternative_endpoints'] = list(self.endpoints_provider)
        if not context['alternative_endpoints']:
            raise ValueError('There isn\'t available endpoint')
        context['endpoint'] = context['alternative_endpoints'].pop(0)

    def should_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt

        Returns
        -------
        bool
        """
        return len(attempt.context['alternative_endpoints']) > 0

    def prepare_retry(self, attempt):
        """
        Parameters
        ----------
        attempt: qiniu.retry.Attempt

        Returns
        -------
        None
        """
        if not attempt.context['alternative_endpoints']:
            raise Exception('There isn\'t available endpoint for next try')
        attempt.context['endpoint'] = attempt.context['alternative_endpoints'].pop(0)
